"""Form components"""
import typing

from textual.containers import Horizontal, Container
from textual.reactive import reactive
from textual.widgets import Label, Select
from nasse.tui.components.inputs import Input
from textual.css.query import NoMatches

from nasse.models import UserSent
from nasse.tui.widget import Widget
from nasse.tui.components.texts import SectionTitle
from nasse.localization import Localization, EnglishLocalization


class UserSentInput(Widget):
    """The input component for a user sent value"""
    value: reactive[typing.Optional[UserSent]] = reactive(None)

    DEFAULT_CSS = """
    UserSentInput {
        height: auto;
        margin: 1 0 1 0;
    }

    .form-input-container {
        height: auto;
    }

    .form-input-name {
        width: 15%;
        height: auto;
    }

    .form-input-value {
        width: 80%;
        height: auto;
    }

    .form-input-description {
        text-opacity: 0.5;
        margin-left: 1;
    }
    """
    input_name: reactive[typing.Optional[str]] = reactive(None)
    input_value: reactive[typing.Optional[str]] = reactive(None)

    def __init__(self,
                 user_sent: typing.Optional[UserSent] = None,
                 inputs: typing.Optional[typing.Set[UserSent]] = None,
                 on_change: typing.Optional[typing.Callable[["UserSentInput", typing.Optional[str], typing.Optional[str]], typing.Any]] = None,
                 initial_value: typing.Optional[str] = None,
                 localization: typing.Type[Localization] = EnglishLocalization,
                 **kwargs) -> None:

        super().__init__(**kwargs)
        self.user_sent = user_sent
        self.inputs = inputs or set()
        self.on_change = on_change
        self.localization = localization

        self.initial_value = initial_value
        if self.user_sent:
            self.input_name = self.user_sent.name

    def compose(self):
        if not self.inputs:
            with Horizontal(classes="form-input-container"):
                yield Input(value=self.user_sent.name if self.user_sent else None, placeholder=self.localization.tui_name, classes="form-input-name", name="input-name")
                yield Input(value=self.initial_value, placeholder=self.localization.tui_value, classes="form-input-value", name="input-value")
            return

        with Horizontal(classes="form-input-container"):
            if not self.user_sent:
                yield Select([(element.name, element) for element in self.inputs], classes="form-input-name", value=None, name="input-name")
                yield Input(disabled=False, classes="form-input-value", value=self.initial_value, name="input-value")
                return
            yield Select([(element.name, element) for element in self.inputs],
                         value=self.user_sent,
                         disabled=self.user_sent.required,
                         classes="form-input-name",
                         name="input-name")
            yield Input(placeholder=self.user_sent.type.__name__
                        if hasattr(self.user_sent.type, "__name__") else str(self.user_sent.type),
                        classes="form-input-value",
                        name="input-value")
        yield Label(self.user_sent.description or "" if self.user_sent else "", classes="form-input-description")

    def on_input_changed(self, event: Input.Changed) -> None:
        """When an input changed"""
        if event.input.name == "input-name":
            self.input_name = event.input.value
        elif event.input.name == "input-value":
            self.input_value = event.input.value

        if self.on_change:
            self.on_change(self, self.input_name, self.input_value)

    def on_select_changed(self, event: Select.Changed) -> None:
        """When a Select object changed"""
        if event.select.name == "input-name":
            val: typing.Optional[UserSent] = event.select.value
            self.input_name = val.name if val else None
            for inp in self.inputs:
                if inp.name == self.input_name:
                    try:
                        self.query_one(".form-input-description", Label).update(inp.description)
                    except NoMatches:
                        self.mount(Label(inp.description or "", classes="form-input-description"))
                    break

        if self.on_change:
            self.on_change(self, self.input_name, self.input_value)


class UserSentForm(Widget):
    """A form for user sent inputs"""

    DEFAULT_CSS = """
    UserSentForm {
        height: auto;
        margin-top: 1;
        margin-bottom: 1;
    }

    .form-inputs-container {
        height: auto;
    }

    .form-buttons {
        height: auto;
        align-horizontal: right;
    }

    .form-buttons-add {
        background: rgb(0, 162, 255);
        color: white;
    }

    .form-buttons-remove {
        opacity: 0.75;
    }
    """

    def __init__(self,
                 title: str,
                 inputs: typing.Optional[typing.Set[UserSent]] = None,
                 multiple: bool = False,
                 initial_values: typing.Optional[typing.List[typing.Tuple[UserSent, str]]] = None,
                 localization: typing.Type[Localization] = EnglishLocalization,
                 **kwargs) -> None:
        super().__init__(**kwargs)
        self.title = title
        self.inputs = inputs or set()
        self.multiple = multiple
        self.localization = localization

        self.initial_values = initial_values or []

    def compose(self):
        yield SectionTitle(self.title)

        initial_values = self.initial_values.copy()

        for element in self.inputs:
            if element.required:
                for user_sent, value in initial_values.copy():
                    if user_sent == element:
                        yield UserSentInput(element, inputs=self.inputs, initial_value=value, localization=self.localization)
                        initial_values.remove((user_sent, value))
                        break
                else:
                    yield UserSentInput(element, inputs=self.inputs, localization=self.localization)

        for user_sent, value in initial_values:
            yield UserSentInput(user_sent, inputs=self.inputs, initial_value=value, localization=self.localization)

        with Container(classes="form-inputs-container"):
            yield UserSentInput(None, self.inputs, self.on_change, localization=self.localization)

    #     with Horizontal(classes="form-buttons"):
    #         yield Button("Add", name="add", classes="form-buttons-add")
    #         yield Button("Remove", name="remove", classes="form-buttons-remove")

    def on_change(self, user_input: UserSentInput, name: typing.Optional[str] = None, value: typing.Optional[str] = None):
        """When something changed in any user input"""
        last_element: UserSentInput = self.query_one(".form-inputs-container", Container).query(UserSentInput).last()

        for element in self.query_one(".form-inputs-container", Container).query(UserSentInput)[:-1]:
            if not element.input_name and not element.input_value:
                element.remove()

        if last_element.input_name or last_element.input_value:
            self.query_one(".form-inputs-container", Container).mount(UserSentInput(None, self.inputs, self.on_change, localization=self.localization))

    @property
    def values(self):
        """Returns the rendered values"""
        results = {}
        # pylint: disable=not-an-iterable
        for element in self.query(UserSentInput):
            # if element.input_name and element.input_value:
            if element.input_name:
                if self.multiple:
                    try:
                        results[element.input_name].append(element.input_value)
                    except KeyError:
                        results[element.input_name] = [element.input_value]
                else:
                    results[element.input_name] = element.input_value
        return results

    # def on_button_pressed(self, event: Button.Pressed) -> None:
    #     """When a button is pressed"""
    #     if event.button.name == "add":
    #         self.query_one(".form-inputs-container", Container).mount(UserSentInput(None, self.inputs))
    #     elif event.button.name == "remove":
    #         self.query_one(".form-inputs-container", Container).query(UserSentInput).last().remove()



"""
A list of internal exceptions for Nasse validations
"""
from nasse.exceptions.base import NasseException


class ValidationError(NasseException):
    """When there is an input validation error"""
    MESSAGE = "Nasse couldn't validate an input"
    EXCEPTION_NAME = "VALIDATION_ERROR"


class ConversionError(ValidationError):
    """An internal error occuring when Nasse can't convert an input to a Nasse object"""
    EXCEPTION_NAME = "INTERNAL_CONVERSION_ERROR"
    MESSAGE = "Nasse couldn't convert a given input to a Nasse object instance"


class MethodsConversionError(ConversionError):
    """An internal error occuring when Nasse can't convert an input to methods"""


class ReturnConversionError(ConversionError):
    """An internal error occuring when Nasse can't convert an input to a Nasse `Return` object"""


class UserSentConversionError(ConversionError):
    """An internal error occuring when Nasse can't convert an input to a Nasse `UserSent` object"""


class ErrorConversionError(ConversionError):
    """An internal error occuring when Nasse can't convert an input to a Nasse `Error` object"""


class LoginConversionError(ConversionError):
    """An internal error occuring when Nasse can't convert an input to a Nasse `Login` object"""


class CookieConversionError(ConversionError):
    """An internal error occuring when Nasse can't convert an input to a Nasse `Cookie` object"""



"""
Provides a lightweight and easy way to access CLI arguments
"""
import sys
import typing


def _enforce_iter(value: typing.Any) -> typing.List[str]:
    """Enforces the value to be a list of strings"""
    if isinstance(value, str) or not isinstance(value, typing.Iterable):
        return [str(value)]
    return [str(val) for val in value]


class Args:
    """A lightweight and easy way to access CLI arguments"""

    @classmethod
    @property
    def args(cls) -> typing.List[str]:
        """A list of CLI arguments passed in when running the program"""
        return sys.argv[1:]

    @classmethod
    def get_multiple(cls, *keys) -> typing.List[str]:
        """Gets all of the values passed in for the given keys"""
        args = _enforce_iter(cls.args)
        future = False
        results = []

        # print(keys)

        for element in args:
            if future:
                # if the previous argument was the key
                results.append(element)
                future = False
            elif element in keys:
                # if we get to the key, the next one should be the argument
                future = True
            else:
                # handling the key=value format
                name, _, value = element.partition("=")
                if name == keys:
                    results.append(value)

        return results

    @classmethod
    def get(cls, *keys, default: typing.Optional[str] = None) -> typing.Optional[str]:
        """Return the value for key if key is in the dictionary, else default"""
        results = cls.get_multiple(*keys)
        if results:
            return results[0]
        return default

    def __class_getitem__(cls, keys) -> str:
        result = cls.get(*_enforce_iter(keys))
        if not result:
            if isinstance(keys, str) or not isinstance(keys, typing.Iterable):
                keys = [str(keys)]
            args = " | ".join(f"`{key}`" for key in keys)
            raise KeyError(f"Did you forget to pass in the {args} argument ?")
        return result

    __getitem__ = __class_getitem__
    __getattr__ = __getitem__

    @classmethod
    def exists(cls, *keys) -> bool:
        """Returns if the given keys exist in the arguments"""
        if cls.get_multiple(*keys):
            return True
        return False

    def __contains__(self, keys) -> bool:
        return self.exists(*_enforce_iter(keys))



"""
This is where the responses and errors are partly processed
"""
import datetime
import dataclasses
import typing

import werkzeug.http
import werkzeug.exceptions

from nasse import config, exceptions, utils


def exception_to_response(value: Exception, config: typing.Optional[config.NasseConfig] = None):
    """
    Internal function to turn an exception to a tuple of values that can be used to make a response
    """
    if isinstance(value, exceptions.NasseException):
        data = value.MESSAGE
        error = value.EXCEPTION_NAME
        code = int(value.STATUS_CODE)
    elif isinstance(value, werkzeug.exceptions.HTTPException):
        code = value.code or 500
        if code >= 500:  # we don't know what kind of exception it might leak
            data = "An error occured on the server while processing your request"
        # we consider that they are fewer non basic exceptions (non 500) that are dangerous to leak (i.e: 4xx errors are related to the client)
        else:
            data = value.description
        error = " ".join(utils.sanitize.split_on_uppercase(value.__class__.__name__)).upper().strip().replace(" ", "_")
    else:
        # converts class names to error names: NasseException -> NASSE_EXCEPTION
        if isinstance(value, type):
            error = " ".join(utils.sanitize.split_on_uppercase(value.__name__)).upper().strip().replace(" ", "_")
        else:
            error = " ".join(utils.sanitize.split_on_uppercase(value.__class__.__name__)).upper().strip().replace(" ", "_")
        if config and config.debug:
            data = "An error occured on the server while processing your request ({error})".format(error=value)
        else:
            data = "An error occured on the server while processing your request"
        code = 500
    if not data:
        data = "An error occured on the server while processing your request"
    return data, error, code


def cookie_validation(value):
    """
    Internal function to validate a value that needs to be a `ResponseCookie` instance
    """
    try:
        if isinstance(value, ResponseCookie):
            return value.__copy__()
        if isinstance(value, str):
            return ResponseCookie(key=value)
        if utils.unpack.is_unpackable(value):
            try:
                return ResponseCookie(**value)
            except TypeError:
                raise exceptions.validate.CookieConversionError(
                    "Either 'name' is missing or one argument doesn't have the right type while creating a Nasse.response.ResponseCookie instance")
        raise ValueError  # will be catched
    except Exception as err:
        if isinstance(err, exceptions.NasseException):
            raise err
        raise exceptions.validate.CookieConversionError(
            "Nasse cannot convert value of type <{t}> to Nasse.response.ResponseCookie".format(t=value.__class__.__name__))


@dataclasses.dataclass
class ResponseCookie:
    """Represents a cookie to send back"""
    key: str
    value: str = ""
    max_age: typing.Optional[typing.Union[float, datetime.timedelta]] = None
    expires: typing.Optional[typing.Union[float, datetime.datetime]] = None
    path: typing.Union[str, tuple, bytes] = "/"
    domain: typing.Optional[typing.Union[str, bytes]] = None
    secure: bool = False
    httponly: bool = False
    samesite = None
    charset: str = "utf-8"
    max_size: int = 4093

    def __post_init__(self):
        self.expires = self.expires or datetime.datetime.now() + datetime.timedelta(days=30)

    def dumps(self) -> str:
        """
        Creates a header value (string) for the given cookie

        Returns
        -------
        str
            The cookie value
        """
        return werkzeug.http.dump_cookie(**dataclasses.asdict(self))


class Response:
    """Represents a response"""

    # Holds the different type arguments
    # Warning: Should only be modified in a copy of the class
    __returning__ = None

    def __class_getitem__(cls, args):
        """
        Called when giving type arguments to the class.

        Parameters
        ----------
        cls: Response
            The actual class which is being instantiated
        args
            The arguments passed with the type

        Example
        -------
        >>> from nasse import Response, Return
        >>> @app.route
        >>> def hello(username: str = "someone") -> Response[Return(example={"hello": "someone"})]:
        ...     return {"hello": username}

        Returns
        -------
        Response
            The instantiated class
        """
        class NewResponse(cls):
            """A subclass containing the type arguments"""
            __returning__ = args if not isinstance(args, tuple) else list(args)

        NewResponse.__name__ = cls.__name__

        return NewResponse

    def __init__(self,
                 data: typing.Any = None,
                 message: typing.Optional[str] = None,
                 error: typing.Optional[str] = None,
                 code: typing.Optional[int] = None,
                 headers: typing.Optional[typing.Dict[str, str]] = None,
                 cookies: typing.Optional[typing.List[ResponseCookie]] = None,
                 content_type: typing.Optional[str] = None) -> None:
        """
        A Response object given to Nasse to format the response

        Parameters
        ----------
        data: typing.Any, default = None
            The data returned to the client
            if 'data' is None, nothing extra is returned to the client
        message: str, default = None
            The message returned to the client
        error: str, default = None
            If an error occured, the error name (i.e the client is not correctly authenticated, you might want to set this as "AUTH_ERROR"
        code: int, default = 200
            The status code to return
        headers: dict[str, str], default = None
            The extra headers to send back (i.e headers=[{"X-NASSE-AUTH": "nasse+1very12897,cool1212798,128129token"}])
        cookies: dict[str, str], default = None
            The cookies to send back
        """
        cookies = cookies or []

        self.data = data

        if isinstance(error, Exception):
            temp_data, error, temp_code = exception_to_response(error)
        else:
            temp_data, temp_code = None, 200

        data = data or temp_data

        code = code or temp_code

        self.error = str(error) if error is not None else None
        self.code = int(code)
        self.message = str(message) if message is not None else None

        self.headers = {}
        if utils.unpack.is_unpackable(headers):
            # headers: {"HEADER-KEY": "Header Value"}
            for header_key, header_value in dict(headers).items():
                self.headers[str(header_key)] = str(header_value)
        elif isinstance(headers, typing.Iterable):
            if isinstance(headers[0], typing.Iterable) and len(headers[0]) == 2:
                # headers: [("HEADER-KEY", "Header Value")]
                for element in headers:
                    self.headers[str(element[0])] = str(element[1])
            elif len(headers) == 2:
                # headers: ("HEADER-KEY", "Header Value")
                self.headers[str(headers[0])] = str(headers[1])

        if content_type:
            self.headers["Content-Type"] = str(content_type)

        self.cookies = []
        if cookies is not None:
            if utils.unpack.is_unpackable(cookies):
                for key, val in dict(cookies).items():
                    item = {"key": str(key)}
                    item.update(val)
                    self.cookies.append(cookie_validation(item))
            elif isinstance(cookies, typing.Iterable):
                for item in cookies:
                    self.cookies.append(cookie_validation(item))
            else:
                self.cookies.append(cookie_validation(cookies))
                raise exceptions.validate.CookieConversionError(
                    "Nasse cannot convert value of type <{t}> to Nasse.response.ResponseCookie".format(t=cookies.__class__.__name__))

    def __copy__(self):
        return Response(
            data=self.data,
            message=self.message,
            error=self.error,
            code=self.code,
            headers=self.headers,
            cookies=self.cookies
        )


def hello() -> Response[{"hello": "world"}]:
    """Hello"""
    return Response()



"""
__main__

The CLI entry
"""

import argparse
import contextlib
import json
import pathlib
import sys
import typing
import os
import importlib.util

import nasse
import nasse.servers
import nasse.localization
from nasse.utils.types import StringEnum
from nasse.tui.apps import http_app


class ServerEnum(StringEnum):
    """The server backend to use"""
    ACCEPTED = ("flask", "gunicorn")
    DEFAULT = "flask"
    LOWER = True


def string_to_server(string: str) -> typing.Type[nasse.servers.ServerBackend]:
    """Retrieves the right server backend to use"""
    string = str(string).strip().lower()
    if string == "gunicorn":
        from nasse.servers.gunicorn import Gunicorn
        return Gunicorn
    from nasse.servers.flask import Flask
    return Flask


def postman_to_endpoints(data: dict):
    """Turns postman data to a collection of Nasse endpoints"""
    results = []
    for element in data.get("item", []):
        req = element.get("request", {})
        url = req.get("url", {}).get("path")
        if not isinstance(url, str):
            url = "/".join(url)
        results.append(nasse.Endpoint(
            path=url,
            name=element.get("name"),
            methods=req.get("method"),
            headers=[
                nasse.Header(name=head.get("key"),
                             description=head.get("description", {}).get("content"))
                for head in req.get("header", [])
            ],
            description=req.get("description"),
            category=data.get("info", {}).get("name")
        ))
    return results


def get_nasse_instance(file: pathlib.Path):
    """Retrieves a Nasse instance from the given file"""
    file = pathlib.Path(file).resolve().absolute()
    with add_to_path(file):
        spec = importlib.util.spec_from_file_location(file.stem, str(file))
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)

        for obj in dir(module):
            element = getattr(module, obj)
            if isinstance(element, nasse.Nasse):
                # YAY
                return element
    raise ValueError("There is no `Nasse` instance in the given file")


def prepare_runnner_parser(parser: argparse.ArgumentParser):
    """Populates the parser with the `runner` app arguments"""
    # parser.add_argument("input", action="store", help="A python file containing the Nasse instance to run")

    parser.add_argument("--host", action="store", type=str, required=False, default=None, help="(server) The host to bind to")
    parser.add_argument("--port", "-p", action="store", type=int, required=False, default=None, help="(server) The port to bind to")
    parser.add_argument("--server", "-s", action="store", type=ServerEnum, required=False, default=ServerEnum.DEFAULT,
                        help="(server) The server to use (accepts: {accepts})".format(accepts=", ".join(ServerEnum.ACCEPTED)))
    parser.add_argument("--watch", "-w", nargs="*", default=["**/*.py"], help="(server) Files to watch changes on debug mode")
    parser.add_argument("--ignore", "-i", nargs="*", default=[], help="(server) Files to ignore when watching for file changes on debug mode")

    parser.add_argument("--debug", "-d", action="store_true", help="(server) To run with debug mode enabled")
    parser.add_argument("--config", "-c", help="(server) A configuration file for extra arguments passed to the server",
                        action="store", type=pathlib.Path, required=False, default=None)


def prepare_docs_parser(parser: argparse.ArgumentParser):
    """Populates the parser with the `docs` app arguments"""
    parser.add_argument("--make-docs", "-md", "--docs", "--generate-docs", action="store_true", help="(docs) Generates the docs and exits")
    parser.add_argument("--language", "--localization", action="store", required=False, help="(docs) The docs language")
    parser.add_argument("--output", "-o", "--docs-dir", "--docs_dir", action="store", required=False, help="(docs) The directory to output the docs")
    parser.add_argument("--docs-curl", action="store_true", help="(docs) If we need to render the curl examples")
    parser.add_argument("--docs-javascript", action="store_true", help="(docs) If we need to render the javascript examples")
    parser.add_argument("--docs-python", action="store_true", help="(docs) If we need to render the python examples")


def get_runner_args(parser: argparse.ArgumentParser) -> typing.Dict[str, typing.Any]:
    """Retrieves the arguments to pass to Nasse"""
    args = parser.parse_args()

    if args.config:
        with open(args.config) as f:
            config = json.load(f)
    else:
        config = {}

    config["debug"] = args.debug

    for attr in ("host", "port", "watch", "ignore"):
        config[attr] = getattr(args, attr)

    config["server"] = string_to_server(args.server)
    return config


def get_docs_args(parser: argparse.ArgumentParser) -> typing.Dict[str, typing.Any]:
    """Retrieves the arguments to pass to Nasse"""
    args = parser.parse_args()

    return {
        "base_dir": args.output or None,
        "curl": args.docs_curl,
        "javascript": args.docs_javascript,
        "python": args.docs_python,
        "localization": http_app.language_to_localization(args.language)
    }


@contextlib.contextmanager
def add_to_path(path: pathlib.Path):
    """
    Parameters
    ----------
    path: pathlib.Path
        the parent path
    """
    old_path = sys.path
    sys.path = sys.path[:]
    sys.path.insert(0, str(path))
    try:
        yield
    finally:
        sys.path = old_path


def main(input: typing.Union[pathlib.Path, str]):
    """The main entrypoint"""
    instance = None
    endpoints = []

    try:
        LOAD_DIRECTORY = bool(int(os.getenv("NASSE_LOAD_DIRECTORY", "0")))
    except Exception:
        LOAD_DIRECTORY = False

    try:
        input = pathlib.Path(input)
        if input.is_file():
            try:
                data = json.loads(input.read_text())
                # Should be a list of endpoints
                # Might be a postman file or a nasse endpoints file
                endpoints.extend(postman_to_endpoints(data))
            except Exception:
                # Should be a python file
                instance = get_nasse_instance(input)

        elif input.is_dir():
            # First searching for common filenames
            for common in ("__init__.py", "app.py", "server.py", "main.py", "run.py"):
                current_input = input / common
                if current_input.is_file():
                    try:
                        instance = get_nasse_instance(current_input)
                    except Exception:
                        pass
            # Then searching through all of the files
            for element in input.iterdir():
                if element.is_file():
                    try:
                        # Might be a JSON file containing Postman data
                        data = json.loads(element.read_text())
                        endpoints.extend(postman_to_endpoints(data))
                        continue
                    except Exception:
                        pass
                    # We won't enable the directory traversing by default
                    # simply because it might load unwanted content
                    if LOAD_DIRECTORY:
                        try:
                            # Assuming this is a Python file
                            instance = get_nasse_instance(element)
                        except Exception:
                            # Welp we were wrong it might not be a file we want to load
                            pass
    except Exception:
        pass

    return instance, endpoints


def entry():
    """The entrypoint for terminals"""
    parser = argparse.ArgumentParser("nasse", description=nasse.__doc__)
    parser.add_argument("--version", "-v", action="version", version=nasse.__version__)
    parser.add_argument("input", action="store", default="", help="The file or URL to use with nasse", nargs="?")

    prepare_runnner_parser(parser)
    prepare_docs_parser(parser)

    args = parser.parse_args()

    instance, endpoints = main(input=args.input)

    if instance:
        if args.make_docs:
            return instance.make_docs(**get_docs_args(parser))
        return instance.run(**get_runner_args(parser))

    if args.make_docs:
        raise ValueError("Couldn't find any Nasse instance to generate the docs fors")

    return http_app.HTTP(str(args.input or "http://localhost"), endpoints=endpoints).run()


if __name__ == "__main__":
    entry()



"""
Manages GitHub flavored markdown headers
"""

import typing


def header_link(header: str, registry: typing.Optional[typing.List[str]] = None) -> str:
    """
    Generates a link from the given header

    Note: All text is converted to lowercase.
    Note: All non-word text (e.g., punctuation, HTML) is removed.
    Note: All spaces are converted to hyphens.
    Note: Two or more hyphens in a row are converted to one.
    Note: If a header with the same ID has already been generated, a unique incrementing number is appended, starting at 1.

    Parameters
    ----------
    header: str
        The header to generate the link from
    registry: list[str], optional
        An internal registry of the different created links

    Returns
    -------
    str
        A link to the header
    """
    if registry is None:
        registry = []
    # registry = registry
    result = "".join(l for l in str(header) if l.isalpha()
                     or l.isdecimal() or l == " ")
    result = result.replace(" ", "-").lower()
    final_result = ""
    for index, letter in enumerate(result):
        if letter == "-" and index > 0 and result[index - 1] != '-':
            final_result += letter
        elif index <= 0 and letter != '-':
            final_result = letter
        elif index > 0 and letter != '-':
            final_result += letter
    link_count = registry.count(final_result)
    registry.append(final_result)
    if link_count > 0:
        return "{result}-{count}".format(result=final_result, count=link_count)
    else:
        return final_result



"""
Automatically generates Markdown Docs

Copyright
---------
Animenosekai
    Original author, MIT License
"""

# pylint: disable=consider-using-f-string

import typing
import inspect
import os.path
import urllib.parse
import pathlib

from nasse import docs, models
from nasse.docs.header import header_link
from nasse.localization.base import Localization
from nasse.utils.sanitize import sort_http_methods


def make_docs(endpoint: models.Endpoint,
              postman: bool = False,
              curl: bool = True,
              javascript: bool = True,
              python: bool = True,
              localization: Localization = Localization(),
              base_dir: typing.Optional[pathlib.Path] = None) -> str:
    """
    Generates the documentation for the given endpoint

    Parameters
    ----------
    endpoint: models.Endpoint
        The endpoint to generate the docs for
    postman: bool, default = False
        If the docs are generated for Postman
        (doesn't include all of the details since a lot is already included by the Postamn docs generator)
    curl: bool, default = True
        If the `curl` examples should be included too
    javascript: bool, default = True
        If the JavaScript examples should be included too
    python: bool, default = True
        If the Python examples should be included too
    localization: Localization, optional
        The language to use to create the docs

    Returns
    -------
    str
        The generated docs
    """

    result = """
# {name}
""".format(name=endpoint.name)

    kwargs = {"endpoint": endpoint,
              "postman": postman,
              "curl": curl,
              "javascript": javascript,
              "python": python,
              "localization": localization,
              "base_dir": base_dir}

    if len(endpoint.methods) <= 1:
        result += """
{}
""".format(endpoint.description.get(list(endpoint.methods)[0], endpoint.description.get("*", localization.no_description)))
        result += make_docs_for_method(**kwargs)
    else:
        for method in sort_http_methods(endpoint.methods):
            kwargs["method"] = method
            result += "\n - ### {}".format(localization.using_method.format(method=method))
            result += "\n{}\n".format(make_docs_for_method(**kwargs))
    return result


def make_docs_for_method(
        endpoint: models.Endpoint,
        method: typing.Optional[str] = None,
        postman: bool = False,
        curl: bool = True,
        javascript: bool = True,
        python: bool = True,
        localization: Localization = Localization(),
        base_dir: typing.Optional[pathlib.Path] = None) -> str:
    """
    Creates the docs for the given method

    Parameters
    ----------
    endpoint: models.Endpoint
        The endpoint to generate the docs for
    method: str, optional
        The method to generate the docs for
    postman: bool, default = False
        If the docs are generated for Postman
        (doesn't include all of the details since a lot is already included by the Postamn docs generator)
    curl: bool, default = True
        If the `curl` examples should be included too
    javascript: bool, default = True
        If the JavaScript examples should be included too
    python: bool, default = True
        If the Python examples should be included too
    localization: Localization, optional
        The language to use to create the docs

    Returns
    -------
    str
        The generated docs
    """
    result = ""

    if len(endpoint.methods) == 1 or method is None:
        # treating it as only a single method endpoint
        heading_level = "###"
        method = list(endpoint.methods)[0]
    else:
        heading_level = "####"
        method = str(method)
        result += """
{}
""".format(endpoint.description.get(method if method in endpoint.description else "*",
                                    localization.no_description))

    try:
        unwrapped = inspect.unwrap(endpoint.handler)
        source_file = inspect.getsourcefile(unwrapped)
    except Exception:
        source_file = endpoint.handler.__code__.co_filename

    try:
        base = pathlib.Path(base_dir or pathlib.Path() / "docs").absolute()
        source_file = pathlib.Path(source_file).absolute()
        common = pathlib.Path(os.path.commonpath([str(source_file), str(base)])).resolve()
        # Getting the relative path
        path = pathlib.Path(source_file).resolve().relative_to(common)
        # Distance between the docs and the most deep common path
        distance = str(base).count("/") - str(common).count("/") + 1
        path = pathlib.Path("../" * distance + str(path))
    except Exception:
        path = pathlib.Path(source_file)

    line = endpoint.handler.__code__.co_firstlineno

    # ENDPOINT HEADING

    if not postman:
        result += """
```http
{method} {path}
```

> [{source_code_path}]({github_path})
""".format(method=method,
           path=endpoint.path,
           source_code_path=path,
           # FIXME: this needs to be fixed because it fails sometimes
           github_path="{path}#L{line}".format(path=path.as_posix(), line=line))

    else:
        result = """
> [{source_code_path}]({github_path})

{description}

""".format(source_code_path=path,
            github_path="../../{path}#L{line}".format(path=path, line=line),
            description=endpoint.description.get(method if method in endpoint.description else "*",
                                                 localization.no_description))


# AUTHENTICATION

    result += """
{heading} {localization__authentication}

""".format(heading=heading_level, localization__authentication=localization.authentication)

    rules = endpoint.login.get(method, endpoint.login["*"])
    if not rules:
        result += localization.no_auth_rule
    else:
        for login_rule in rules:
            if login_rule.skip:
                result += localization.no_login
            else:
                if login_rule.required:
                    if len(login_rule.types) > 0:
                        result += localization.login_with_types_required.format(types=', '.join(str(type_name) for type_name in login_rule.types))
                    else:
                        result += localization.login_required
                else:
                    if len(login_rule.types) > 0:
                        result += localization.login_with_types_optional.format(types=', '.join(str(type_name) for type_name in login_rule.types))
                    else:
                        result += localization.login_optional

            if login_rule.skip_fetch:
                result += localization.login_suffix_only_verified

    if not postman:  # POSTMAN DOESN'T NEED THESE INFORMATION

        # USER SENT VALUES

        for field, values in [
            (localization.dynamic_url, endpoint.dynamics),
            (localization.parameters, endpoint.params),
            (localization.headers, endpoint.headers),
            (localization.cookies, endpoint.cookies)
        ]:
            params = models.get_method_variant(method, values)
            if len(params) > 0:
                result += """

{heading} {field}

| {localization__name}         | {localization__description}                      | {localization__required}         | {localization__type}             |
| ------------ | -------------------------------- | ---------------- | ---------------- |
""".format(field=field,
                    heading=heading_level,
                    localization__name=localization.name,
                    localization__description=localization.description,
                    localization__required=localization.required,
                    localization__type=localization.type)
                result += "\n".join(
                    ["| `{param}` | {description}  | {required}            | {type}            |".format(param=param.name,
                                                                                                         description=param.description or localization.no_description,
                                                                                                         required=localization.yes if param.required else localization.no,
                                                                                                         type=param.type.__name__ if hasattr(param.type, "__name__") else str(param.type) if param.type is not None else "str") for param in params])


# LANGUAGE SPECIFIC EXAMPLES

        if any((curl, javascript, python)):
            result += """

{heading} {localization__example}

<!-- tabs:start -->
""".format(heading=heading_level, localization__example=localization.example)
            for proceed, highlight, language, function in [
                (curl, "bash", "cURL", docs.curl.create_curl_example_for_method),
                (javascript, "javascript", "JavaScript", docs.javascript.create_javascript_example_for_method),
                (python, "python", "Python", docs.python.create_python_example_for_method)
            ]:
                if proceed:
                    result += """

<details>
    <summary>{language} {localization__example}</summary>

{heading} **{language}**

```{highlight}
{example}
```

</details>
""".format(heading=heading_level + "#",
                        localization__example=localization.example,
                        highlight=highlight,
                        language=language,
                        example=function(endpoint, method=method))
            result += """<!-- tabs:end -->"""


# RESPONSE
    returning = models.get_method_variant(method, endpoint.returns)
    if len(returning) > 0:
        result += """

{heading} {localization__response}
""".format(heading=heading_level, localization__response=localization.response)


# JSON RESPONSE EXAMPLE

        if endpoint.json:
            result += """
{heading} {localization__example_response}

```json
{example}
```
""".format(heading=heading_level + "#",
                localization__example_response=localization.example_response,
                example=docs.example.generate_example(endpoint, method=method))
        else:
            result += "\n"
            result += localization.not_json_response
            result += "\n"

# RESPONSE DESCRIPTION

        result += """
{heading} {localization__returns}

| {localization__field}        | {localization__description}                      | {localization__type}   | {localization__nullable}  |
| ----------   | -------------------------------- | ------ | --------- |
""".format(heading=heading_level + "#",
           localization__returns=localization.returns,
           localization__field=localization.field,
           localization__description=localization.description,
           localization__type=localization.type,
           localization__nullable=localization.nullable)

        result += "\n".join(["| `{key}` | {description}  | {type}      | {nullable}      |".format(key=element.name,
                                                                                                   description=element.description or localization.no_description,
                                                                                                   type=docs.example._get_type(element),
                                                                                                   nullable=localization.yes if element.nullable else localization.no)
                             for element in returning])

    result += "\n"


# ERRORS

    errors = models.get_method_variant(method, endpoint.errors)
    if len(errors) > 0:
        result += """
{heading} {localization__possible_errors}

| {localization__exception}         | {localization__description}                      | {localization__code}   |
| ---------------   | -------------------------------- | ------ |
""".format(heading=heading_level + "#",
           localization__possible_errors=localization.possible_errors,
           localization__exception=localization.exception,
           localization__description=localization.description,
           localization__code=localization.code)

        result += "\n".join(["| `{key}` | {description}  | {code}  |".format(key=error.name,
                                                                             description=error.description or localization.no_description,
                                                                             code=error.code)
                             for error in errors])


# INDEX LINKING

    result += "\n[{localization__return_to_index}](../{localization__getting_started}.md#{localization__index})".format(
        localization__return_to_index=localization.return_to_index,
        localization__getting_started=urllib.parse.quote(localization.getting_started, safe=""),
        localization__index=header_link(localization.index))
    return result



"""
Unpack utils
"""
import typing


def is_unpackable(obj: typing.Any):
    """
    Checks if the given object is unpackable or not (if you can use **obj or not)
    """
    return all(hasattr(obj, attr) for attr in ('keys', '__getitem__'))



"""
File containing the different classes used in Nasse
"""
import abc
import dataclasses
import inspect
import pathlib
import typing
import miko
from nasse import utils, response


class Types:
    """Holds different types"""
    # Type Aliases
    T = typing.TypeVar("T")

    class Method:
        """HTTP method types"""
        Standard = typing.Literal["GET", "HEAD", "POST",
                                  "PUT", "DELETE", "CONNECT",
                                  "OPTIONS", "TRACE", "PATCH"]
        Any = typing.Union[Standard, str]

    Type = typing.Union[str, typing.Callable[[str], typing.Any], typing.Type]
    HandlerOutput = typing.Union["response.Response", Exception, typing.Iterable, typing.Generator]

    FinalMethodVariant = typing.Dict[Method.Any, T]
    FinalIterable = typing.Set[T]

    MethodVariant = typing.Optional[typing.Union[FinalMethodVariant[T], T]]
    OptionalIterable = typing.Optional[typing.Union[typing.Iterable[T], T]]


# Validations


def method_validation(method: typing.Any):
    """Validates the given method"""
    result = str(method).upper().strip()
    if result not in typing.get_args(Types.Method.Standard) and result != "*":
        utils.logging.logger.warn(f"Defining non standard HTTP method {result}")
    return result


def path_to_name(path: str):
    """Turns a path into an endpoint name"""
    return " > ".join(elem
                      .title()
                      .replace("-", " ")
                      .replace("_", " ")
                      for elem in str(path).split("/"))


def get_method_variant(method: str,
                       value: Types.FinalMethodVariant[Types.FinalIterable[Types.T]]) -> Types.FinalIterable[Types.T]:
    """Returns the variant defined for the given method"""
    return value.get("*", set()).union(value.get(method, set()))


def complete_cast(value: typing.Any, cast: typing.Type[Types.T], iter: bool = False) -> Types.T:
    """Casts the given value with the given type"""
    if iter:
        return validates_optional_iterable(value, cast)
    try:
        if isinstance(value, cast):
            return value
    except TypeError:
        pass
    if dataclasses.is_dataclass(cast):
        if dataclasses.is_dataclass(value):
            return cast(**dataclasses.asdict(value))
        else:
            return cast(**value)
    return cast(value)


@typing.overload
def validates_method_variant(value: Types.MethodVariant[Types.T],
                             cast: typing.Type[Types.T],
                             iter: bool = True) -> typing.Dict[Types.Method.Any, Types.FinalIterable[Types.T]]: ...


@typing.overload
def validates_method_variant(value: Types.MethodVariant[Types.T],
                             cast: typing.Type[Types.T],
                             iter: bool = False) -> typing.Dict[Types.Method.Any, Types.T]: ...


def validates_method_variant(value, cast, iter=False):
    """Validates a value which might vary with the method"""
    if not value:
        if iter:
            return {"*": set()}
        else:
            return {"*": None}
    try:
        return {method_validation(key): complete_cast(value, cast, iter)
                for key, value in {**value}.items()}
    except TypeError:
        return {"*": complete_cast(value, cast, iter)}


def validates_optional_iterable(value: Types.OptionalIterable[Types.T], cast: typing.Type[Types.T]) -> typing.Set[Types.T]:
    """Validates an iterable which might be None"""
    if not value:
        return set()
    if isinstance(value, str) or not isinstance(value, typing.Iterable):
        return {complete_cast(value, cast, iter=False)}
    return {complete_cast(val, cast, iter=False) for val in value}


# Endpoint Shaping Models


@dataclasses.dataclass(eq=True, frozen=True)
class Return():
    """A return value"""
    name: str
    """The name of the field"""
    example: typing.Any = None
    """An example of returned value"""
    description: typing.Optional[str] = None
    """A description of the returned value"""
    type: Types.Type = str
    """The type of returned value"""
    children: typing.List["Return"] = dataclasses.field(default_factory=list)
    """The different children values"""
    nullable: bool = False
    """If this value is can be null (None)"""

    def __hash__(self) -> int:
        return hash(self.name) + hash(self.description) + hash(self.type) + hash(self.nullable)


def init_class(cls: typing.Type[Types.T], instance: Types.T, **kwargs):
    """Initialize a class"""
    if hasattr(cls, "__annotations__"):
        for attr in cls.__annotations__:
            if attr in kwargs:
                setattr(instance, attr, kwargs[attr])
            # else:
            #     setattr(instance, attr, None)


@dataclasses.dataclass(eq=True)
class Login:
    """Defines the rules for the login methods"""
    required: bool = False
    """If the login is required or not. The user may still authenticate."""
    types: typing.Set[typing.Any] = None
    """Accepted types of accounts"""
    skip: bool = False
    """Whether to completely skip or not the authentication step"""
    skip_fetch: bool = False
    """Whether to skip fetching the account or not.
    This effectively only checks if the provided token is correct or not."""

    def __init__(self,
                 required: bool = False,
                 types: typing.Optional[typing.Union[typing.Iterable[typing.Any],
                                                     typing.Any]] = None,
                 skip: bool = False,
                 skip_fetch: bool = False):

        init_class(Login, self,
                   required=required,
                   types=types,
                   skip=skip,
                   skip_fetch=skip_fetch)

        # Would have used __post_init__ but it doesn't support
        # having different __init__ and type annotations
        if not self.types:
            self.types = set()
        elif isinstance(self.types, str) or not isinstance(self.types, typing.Iterable):
            self.types = {self.types}
        else:
            self.types = set(self.types)

    def __hash__(self) -> int:
        result = 0
        for attr in ("required", "skip", "skip_fetch"):
            element = getattr(self, attr)
            result += hash(element)
        for element in self.types:
            result += hash(element)

        return result


@dataclasses.dataclass(eq=True, frozen=True)
class UserSent:
    """A value sent by the user"""
    name: str
    """The name of the value sent"""
    description: typing.Optional[str] = None
    """A description of the value sent"""
    required: bool = True
    """If the value is required or not"""
    type: Types.Type = str
    """The type of value sent by the user"""


Dynamic = UserSent
"""A dynamic path component"""
Header = UserSent
"""A header"""
Parameter = UserSent
"""A query parameter"""
Cookie = UserSent
"""A cookie"""

# @dataclasses.dataclass
# class Dynamic(UserSent):
#     """A dynamic path component"""


# @dataclasses.dataclass
# class Header(UserSent):
#     """A header"""


# @dataclasses.dataclass
# class Parameter(UserSent):
#     """A query parameter"""


# @dataclasses.dataclass
# class Cookie(UserSent):
#     """A cookie"""

# Backward compatibility
Param = Parameter


@dataclasses.dataclass(eq=True, frozen=True)
class Error:
    """An error to be raised when something goes wrong"""
    name: str
    """The name of the error"""
    description: typing.Optional[str] = None
    """A description of a situation where this error might be raised"""
    code: int = 500
    """The status code of the response sent along this error"""


def non_implemented():
    """This represents a non implemented endpoint"""
    return NotImplementedError("Unitialized Endpoint")


@dataclasses.dataclass
class Endpoint:
    """Represents an endpoint"""
    handler: typing.Callable[..., Types.HandlerOutput]
    """The function which will handle the request"""
    name: str
    """The name of the endpoint"""
    category: str
    """The category the endpoint is in"""
    sub_category: str
    """The sub category the endpoint is in"""
    description: Types.FinalMethodVariant[str]
    """A description of what the endpoint does"""
    base_dir: pathlib.Path
    """The base directory of the endpoints,
    to define endpoints in a sub-directory without
    altering the path definition"""

    # Request
    path: str
    """The path to the endpoint"""
    methods: Types.FinalIterable[Types.Method.Any]
    """The methods allowed on this endpoint"""
    login: Types.FinalMethodVariant[Types.FinalIterable[Login]]
    """The login rules for this endpoint.
    Defines who has access to this endpoint."""

    # User Sent
    parameters: Types.FinalMethodVariant[Types.FinalIterable[Parameter]]
    """The required and accepted parameters for this endpoint"""
    headers: Types.FinalMethodVariant[Types.FinalIterable[Header]]
    """The required and accepted headers for this endpoint"""
    cookies: Types.FinalMethodVariant[Types.FinalIterable[Cookie]]
    """The required and accepted cookies for this endpoint"""
    dynamics: Types.FinalMethodVariant[Types.FinalIterable[Dynamic]]
    """The required and accepted dynamic parts of the URL for this endpoint"""

    # Response
    json: bool
    """Whether the returned response should be JSON formatted or not"""
    returns: Types.FinalMethodVariant[Types.FinalIterable[Return]]
    """The structure of the returned value"""
    errors: Types.FinalMethodVariant[Types.FinalIterable[Error]]
    """The errors which can be raised by the endpoint"""

    @property
    def params(self):
        """An alias for `parameters`"""
        return self.parameters

    def __init__(self,
                 handler: typing.Callable[..., Types.HandlerOutput] = non_implemented,
                 name: str = "",
                 category: typing.Optional[str] = "Main",
                 sub_category: str = "",
                 description: Types.MethodVariant[str] = None,
                 base_dir: typing.Union[pathlib.Path, str, None] = None,
                 endpoint: typing.Union["Endpoint", typing.Mapping, None] = None,

                 # Request,
                 path: typing.Optional[str] = None,
                 methods: Types.OptionalIterable[Types.Method.Any] = "*",
                 login: Types.MethodVariant[Types.OptionalIterable[Login]] = None,

                 # User Sent,
                 parameters: Types.MethodVariant[Types.OptionalIterable[Parameter]] = None,
                 headers: Types.MethodVariant[Types.OptionalIterable[Header]] = None,
                 cookies: Types.MethodVariant[Types.OptionalIterable[Cookie]] = None,
                 dynamics: Types.MethodVariant[Types.OptionalIterable[Dynamic]] = None,

                 # Response,
                 json: bool = True,
                 returns: Types.MethodVariant[Types.OptionalIterable[Return]] = None,
                 errors: Types.MethodVariant[Types.OptionalIterable[Error]] = None):

        # Merging with `endpoint`
        # Could use **kwargs but it would lose the typings for type-checkers

        initial = {
            "handler": handler,
            "name": name,
            "category": category,
            "sub_category": sub_category,
            "description": description,
            "base_dir": base_dir,
            # "endpoint": endpoint,
            "path": path,
            "methods": methods,
            "login": login,
            "parameters": parameters,
            "headers": headers,
            "cookies": cookies,
            "dynamics": dynamics,
            "json": json,
            "returns": returns,
            "errors": errors
        }

        # Getting the file where the function got defined
        # module = inspect.getmodule(self.handler)
        # if module:
        #     filepath = pathlib.Path(module.__file__)
        # else:
        try:
            filepath = inspect.getsourcefile(inspect.unwrap(handler))
            if not filepath:
                raise ValueError("internal: filepath cannot be None")
        except Exception:
            filepath = handler.__code__.co_filename

        filepath = pathlib.Path(filepath or "")

        # Getting the handler signature
        signature = inspect.signature(handler)

        # Parsing the doc-string
        docs = miko.Docs(handler.__doc__ or "", signature)

        if not name:
            initial["name"] = handler.__name__

        # I might add custom parsers for each method
        if not description:
            initial["description"] = docs.description

        init_args = {k: v for k, v in initial.items() if v}

        try:
            extra_args = dataclasses.asdict(endpoint)
        except Exception:
            try:
                extra_args = {**endpoint}
            except Exception:
                extra_args = {}

        for key, value in extra_args.items():
            if key == "path":
                # `path` would seem implemented by `endpoint` but it isn't
                continue
            init_args.setdefault(key, value)

        for key, value in initial.items():
            init_args.setdefault(key, value)

        if not init_args["category"]:
            init_args["category"] = (filepath.stem or
                (handler.__module__ or "").rpartition(".")[2] or
                "Main")

        # init_args = {k: v for k, v in init_args.items() if v}

        # Initializing instance
        init_class(Endpoint, self, **init_args)

        # Type Validations
        self.description = validates_method_variant(self.description, str)
        self.base_dir = pathlib.Path(self.base_dir) if self.base_dir else pathlib.Path()

        if not self.path:
            if self.base_dir:
                # Validates the base directory
                self.base_dir = pathlib.Path(self.base_dir).resolve().absolute()

                # Temp variables to manipulate the base path
                base = self.base_dir.as_posix()
                base_len = len(base)

                # A fail-safe version of pathlib.Path.relative_to
                result = ""
                # removing the suffix
                for index, letter in enumerate(filepath.resolve().absolute().as_posix().rpartition(".")[0]):
                    # If we are still within the base path
                    # And the letter is in the base path
                    if index < base_len and letter == base[index]:
                        continue
                    result += letter
                self.path = (utils.sanitize.to_path(result)
                             + utils.sanitize.to_path(self.handler.__name__))
            else:
                # it should never come here (?)
                # it was a part used before which shouldn't be ran because we never set `base_dir` to None
                try:
                    self.path = (utils.sanitize.to_path(inspect.getfile(self.handler)) +
                                 utils.sanitize.to_path(self.handler.__name__))
                except Exception:
                    module = inspect.getmodule(self.handler)
                    if module:
                        if module.__name__ == "__main__" and module.__file__:
                            name = pathlib.Path(module.__file__).stem
                        else:
                            name = module.__name__
                    else:
                        name = ""
                    self.path = (utils.sanitize.to_path(name)
                                 + utils.sanitize.to_path(self.handler.__name__))

        self.methods = validates_optional_iterable(self.methods, method_validation)
        self.login = validates_method_variant(self.login, Login, iter=True)

        parsed_path = utils.router.Path(self.path)
        self.path = parsed_path.join(flask=True)

        self.parameters = validates_method_variant(self.parameters, Parameter, iter=True)
        self.dynamics = validates_method_variant(self.dynamics, Dynamic, iter=True)

        # retrieving all of the already defined parameters
        names = ["app", "nasse", "config", "logger", "endpoint",
                 "nasse_endpoint", "request", "method", "values",
                 "params", "parameters", "args", "form", "headers",
                 "account", "dynamics"]

        param_names = []
        for parameters in self.parameters.values():
            param_names.extend([param.name for param in parameters])

        dyn_names = []
        for dynamics in self.dynamics.values():
            dyn_names.extend([dynamic.name for dynamic in dynamics])

        names.extend(param_names)
        names.extend(dyn_names)

        names = set(names)

        # checking all of the dynamic parameters of the path
        for dynamic in parsed_path.dynamics:
            if not dynamic.name in dyn_names:
                # adding the dynamic parameter of the path to the endpoint definition
                if dynamic.name in docs.parameters.elements:
                    parameter = docs.parameters.elements[dynamic.name]

                    element = Dynamic(dynamic.name,
                                      description=parameter.body,
                                      required=not parameter.optional,
                                      type=next(iter(parameter.types)) if parameter.types else dynamic.cast)
                else:
                    element = Dynamic(dynamic.name,
                                      type=dynamic.cast)

                try:
                    self.dynamics["*"].add(element)
                except KeyError:
                    self.dynamics["*"] = {element}
                dyn_names.append(element.name)
                names.add(element.name)

        # checking the parameters defined at the function definition level
        for parameter in docs.parameters.elements.values():
            # from miko.parser.list import Parameter
            # parameter: Parameter
            if not parameter.name in names:
                # adding the parameter if it is a function argument
                element = Parameter(parameter.name,
                                    description=parameter.body,
                                    required=not parameter.optional,
                                    type=next(iter(parameter.types)) if parameter.types else None)
                try:
                    self.parameters["*"].add(element)
                except KeyError:
                    self.parameters["*"] = {element}

        self.headers = validates_method_variant(self.headers, Header, iter=True)
        self.cookies = validates_method_variant(self.cookies, Cookie, iter=True)

        try:
            handler_return = signature.return_annotation
            if not returns and issubclass(handler_return, response.Response):
                self.returns = handler_return.__returning__
        except TypeError:  # issubclass
            pass

        self.returns = validates_method_variant(self.returns, Return, iter=True)
        self.errors = validates_method_variant(self.errors, Error, iter=True)

    def __getitem__(self, key: str):
        return getattr(self, key)


class AccountManagement(abc.ABC):
    """
    An object to verify accounts used by Nasse to determine wether a request is correctly authenticated
    """
    @abc.abstractmethod
    def retrieve_type(self, account):
        """
        An abstract method (that needs to be replaced) to retrieve the account type from a custom Account object
        """
        return None

    @abc.abstractmethod
    def retrieve_account(self, token: str):
        """
        An abstract method (that needs to be replaced) to retrieve an Account object from a token sent by the client
        """
        return None

    @abc.abstractmethod
    def verify_token(self, token: str):
        """
        An abstract method (that needs to be replaced) to verify a token sent by the client (to avoid retrieving the account while still checking for the token)
        """
        return None



"""
Makes HTTP requests

        ╭────────────┬──────────────────────────────────────┬────────────╮
        │  History   |              Request                 |  Explorer  │
        │            | ▒▒ /path                             |            │
        │ /yay       |                                      | Reset      │
        │ POST 200   | Parameters                           | Test       │
        │            | ▒▒▒▒ ▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓      | ╰── Sub Cat│
        │ /no        | ▒▒▒▒ ▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓      |      ╰── So│
        │ ERROR      |                                      |            │
        │            | Headers                              |            │
        │ /exist     ├──────────────────────────────────────┤            │
        │ GET 404    |               Result                 |            │
        │            | POST /yay                            |            │
        │ 4 / 8 / 62 | 200 OK in 469ms                      |            │
        │▆▆█▁▁▄▄▆▆▁▁▁|                                      |            │
        ╰────────────┴──────────────────────────────────────┴────────────╯
"""
# TODO
# - Profiles
import sys
import dataclasses
import pathlib
import time
import typing
import urllib.parse as url

import requests
from rich.traceback import Traceback
from textual import work
from textual.binding import Binding
from textual.containers import Container, Horizontal, VerticalScroll
from textual.events import Click
from textual.reactive import reactive, var
from textual.suggester import Suggester
from textual.validation import Integer, Number
from textual.widgets import (Button, Footer, Header, Label,
                             LoadingIndicator, Pretty, Select, Static, Switch, Markdown,
                             Tree, _header)
from nasse.tui.components.inputs import Input
from textual.worker import get_current_worker

from nasse import __info__
from nasse.localization import EnglishLocalization, Localization, FrenchLocalization, JapaneseLocalization
from nasse.models import Endpoint, Types, UserSent, get_method_variant
from nasse.tui.app import App
from nasse.tui.components import series
from nasse.tui.components.forms import UserSentForm
from nasse.tui.components.headers import StickyHeader
from nasse.tui.components.history import HistoryResponse
from nasse.tui.components.texts import SectionTitle
from nasse.tui.screens import FileBrowser, OptionsScreen, QuitScreen
from nasse.tui.widget import Widget

# pylint: disable=pointless-string-statement
"""
╭─────────────────────────────── Dataclasses ────────────────────────────────╮
│ HTTPOptions                                                                │
│ Loading                                                                    │
│ Error                                                                      │
╰────────────────────────────────────────────────────────────────────────────╯
"""


def language_to_localization(lang: str = "eng"):
    """Returns the correct localization from the given language string"""
    if lang == JapaneseLocalization.__id__:
        return JapaneseLocalization
    elif lang == FrenchLocalization.__id__:
        return FrenchLocalization
    return EnglishLocalization

# @dataclasses.dataclass
# class Profile:
#     """A profile"""
#     name: str
#     parameters: typing.Dict[str, typing.List[str]] = dataclasses.field(default_factory=dict)
#     headers: typing.Dict[str, str] = dataclasses.field(default_factory=dict)
#     cookies: typing.Dict[str, str] = dataclasses.field(default_factory=dict)


@dataclasses.dataclass
class HTTPOptions:
    """App options"""
    language: str = "eng"
    timeout: float = 10
    allow_redirects: bool = True
    proxies: typing.Dict[str, str] = dataclasses.field(default_factory=dict)
    verify: bool = True
    cert: typing.List[str] = dataclasses.field(default_factory=list)
    history_limit: int = 10
    endpoints_update: int = 60
    # profiles: typing.List[Profile] = dataclasses.field(default_factory=list)
    # stream
    # hooks


@dataclasses.dataclass
class Loading:
    """The request loading state"""
    url: str


@dataclasses.dataclass
class Error:
    """An error"""
    exception: Exception
    method: str
    url: str
    params: typing.Dict[str, list] = dataclasses.field(default_factory=dict)
    headers: typing.Dict[str, str] = dataclasses.field(default_factory=dict)
    cookies: typing.Dict[str, str] = dataclasses.field(default_factory=dict)
    files: typing.List[typing.Tuple[str, str]] = dataclasses.field(default_factory=list)
    data: typing.Optional[bytes] = None

    timeout: float = 10
    allow_redirects: bool = True
    verify: bool = True
    proxies: typing.Dict[str, str] = dataclasses.field(default_factory=dict)
    cert: typing.List[str] = dataclasses.field(default_factory=list)


# pylint: disable=pointless-string-statement
"""
╭─────────────────────────────── Widgets/Views/Screens ────────────────────────────────╮
│ View                                                                                 │
│ FileInput                                                                            │
│ OptionsScreen                                                                        │
╰──────────────────────────────────────────────────────────────────────────────────────╯
"""


class View(VerticalScroll):
    """A view in the main screen"""

    def __init__(self, on_click: typing.Callable[[Click], typing.Any], **kwargs) -> None:
        super().__init__(**kwargs)
        self.when_clicked = on_click

    def on_click(self, event: Click):
        return self.when_clicked(event)


class FileInput(Widget):
    """Represents a file"""

    DEFAULT_CSS = """
    FileInput {
        width: auto;
        height: auto;
    }

    .file-input-container {
        height: auto;
    }

    .file-name {
        height: auto;
        width: 40w;
    }

    .file-input {
        height: auto;
        width: 40w;
    }

    .file-input-full {
        width: 80w;
    }

    .file-delete {
        height: auto;
        width: 10w;
    }
    """

    def __init__(self,
                 file: pathlib.Path,
                 on_delete: typing.Optional[typing.Callable[[Widget, pathlib.Path], typing.Any]] = None,
                 prompt_name: bool = True,
                 localization: typing.Type[Localization] = EnglishLocalization,
                 **kwargs):
        self.file = pathlib.Path(file)
        self.on_delete = on_delete
        self.prompt_name = prompt_name
        self.localization = localization
        super().__init__(**kwargs)

    def compose(self):
        elements = []
        if self.prompt_name:
            elements.append(Input(placeholder=self.localization.tui_name, classes="file-name"))
        file_input = Input(str(self.file.resolve()), disabled=True, classes="file-input")
        if not self.prompt_name:
            file_input.add_class("file-input-full")
        elements.append(file_input)
        elements.append(Button("Delete", "error", classes="file-delete"))
        yield Horizontal(*elements, classes="file-input-container")

    @property
    def input_name(self):
        """The file input name"""
        return self.query_one(".file-name", Input).value

    def on_button_pressed(self, event):
        """When a button is pressed"""
        if self.on_delete:
            self.on_delete(self, self.file)


class HTTPOptionsScreen(OptionsScreen[HTTPOptions]):
    """The HTTP app options screen"""

    def __init__(self,
                 base_url: str,
                 on_url_change: typing.Callable[[str], typing.Any],
                 localization: typing.Type[Localization] = EnglishLocalization,
                 **kwargs) -> None:
        self.base_url = str(base_url)
        self.on_url_change = on_url_change
        self.localization = localization
        super().__init__(**kwargs)

    def compose_options(self):
        """Composes the inner options view"""
        with VerticalScroll(id="options-inner-container"):
            yield SectionTitle(self.localization.tui_language)
            yield Select([(local.__native__, local.__id__) for local in (EnglishLocalization, FrenchLocalization, JapaneseLocalization)],
                         prompt=self.localization.tui_language,
                         allow_blank=False,
                         value=self.localization.__id__,
                         id="options-language")
            yield Label(self.localization.tui_language_notice, id="options-language-notice")

            yield SectionTitle(self.localization.tui_base_url)
            yield Input(self.base_url,
                        placeholder=self.localization.tui_base_url_placeholder,
                        id="options-base-url")

            yield SectionTitle(self.localization.tui_endpoints_update)
            yield Input(str(self.options.endpoints_update),
                        placeholder=self.localization.tui_endpoints_update_placeholder,
                        validators=[Integer(minimum=0, failure_description="The time must be a positive integer")],
                        id="options-endpoints-update")

            yield SectionTitle(self.localization.tui_history_limit)
            yield Input(str(self.options.history_limit),
                        placeholder=self.localization.tui_history_limit_placeholder,
                        validators=[Integer(minimum=0, failure_description="The limit must be a positive integer")],
                        id="options-history-limit")

            yield SectionTitle(self.localization.tui_timeout)
            yield Input(str(self.options.timeout),
                        placeholder=self.localization.tui_timeout_placeholder,
                        validators=[Number(minimum=0.0, failure_description="The timeout must be a positive number")],
                        id="options-timeout")

            yield SectionTitle(self.localization.tui_redirects)
            yield Horizontal(
                Label(self.localization.tui_allow_redirects, id="options-redirects-title", classes="options-switch-title"),
                Switch(value=self.options.allow_redirects, id="options-redirects-switch"),
                id="options-redirects-container",
                classes="options-switch-container"
            )

            yield UserSentForm(self.localization.tui_proxies, id="options-proxies", initial_values=[(UserSent(name=key), value) for key, value in self.options.proxies.items()], localization=self.localization)

            yield SectionTitle(self.localization.tui_security)
            yield Horizontal(
                Label(self.localization.tui_verify_request, id="options-verify-title", classes="options-switch-title"),
                Switch(value=self.options.verify, id="options-verify-switch"),
                id="options-verify-container",
                classes="options-switch-container"
            )

            with Container(id="certificate-files", classes="files-input-container"):
                with Container(id="certificate-files-container", classes="files-container"):
                    for file in self.options.cert:
                        yield FileInput(pathlib.Path(file), self.delete_cert_file, prompt_name=False, localization=self.localization)
                yield Button(self.localization.tui_add_certificate, id="add-certificate-button", classes="add-file-button")

    def collect_values(self) -> typing.Dict[str, typing.Any]:
        """Collect the different options value"""
        new_url = self.query_one("#options-base-url", Input).value
        if new_url != self.base_url:
            self.on_url_change(new_url)

        return {
            "language": self.query_one("#options-language", Select).value,
            "endpoints_update": int(self.query_one("#options-endpoints-update", Input).value),
            "history_limit": int(self.query_one("#options-history-limit", Input).value),
            "timeout": float(self.query_one("#options-timeout", Input).value),
            "allow_redirects": self.query_one("#options-redirects-switch", Switch).value,
            "proxies": self.query_one("#options-proxies", UserSentForm).values,
            "verify": self.query_one("#options-verify-switch", Switch).value,
            "cert": [str(inp.file) for inp in self.query(FileInput)]  # pylint: disable=not-an-iterable
        }

    def on_button_pressed(self, event: Button.Pressed) -> None:
        """When a button is pressed"""
        event.stop()
        if event.button.has_class("add-file-button"):
            self.app.push_screen(FileBrowser(), self.add_cert_file)

    def add_cert_file(self, file: typing.Optional[pathlib.Path] = None):
        """Adds a cert file"""
        if file:
            self.query_one("#certificate-files-container", Container).mount(FileInput(file, self.delete_cert_file, prompt_name=False))

    def delete_cert_file(self, file_input: FileInput, file: pathlib.Path):
        """When a file is removed from the files list"""
        file_input.remove()


# pylint: disable=pointless-string-statement
"""
╭─────────────────────────────── App ────────────────────────────────╮
│ HTTP                                                               │
╰────────────────────────────────────────────────────────────────────╯
"""


def define_bindings(localization: typing.Type[Localization] = EnglishLocalization):
    """Defines the different bindings for the app"""
    return [("h", "toggle_history", localization.tui_history), ("r", "toggle_results", localization.tui_result), ("e", "toggle_explorer", localization.tui_explorer),
            ("s", "submit", localization.tui_submit), ("o", "open_options", localization.tui_options), ("q", "request_quit", localization.tui_quit), Binding("ctrl+c", "request_quit", localization.tui_quit, show=False)]


class HTTP(App):
    """Lets you make HTTP requests comfortably"""

    # Default values
    CSS_PATH = "../styles/http.css"
    BINDINGS = define_bindings()

    # Atrributes
    toggle_history = var(True)
    toggle_results = var(False)
    toggle_explorer = var(False)
    history: reactive[typing.List[typing.Union[requests.Response, Error]]] = reactive(list)
    result: reactive[typing.Optional[typing.Union[requests.Response, Error, Loading]]] = reactive(None)
    endpoint: reactive[typing.Optional[Endpoint]] = reactive(None)
    endpoints: reactive[typing.List[Endpoint]] = reactive(list)
    _method: reactive[Types.Method.Any] = reactive("*")
    # profile: reactive[str] = reactive("Default")

    @property
    def method(self) -> str:
        """Get the currently selected method"""
        if self._method == "*":
            try:
                return next(iter(self.endpoint.methods))
            except Exception:
                return "*"
        else:
            return self._method

    @property
    def categories(self) -> typing.Dict[str, typing.Dict[str, typing.List[Endpoint]]]:
        """Returns the category separated endpoints"""
        mid_results: typing.Dict[str, typing.List[Endpoint]] = {}
        for endpoint in self.endpoints:
            try:
                mid_results[endpoint.category].append(endpoint)
            except KeyError:
                mid_results[endpoint.category] = [endpoint]

        results = {}
        for category, endpoints in mid_results.items():
            results[category] = {}
            for endpoint in endpoints:
                try:
                    results[category][endpoint.sub_category or "@TopLevelEndpoint"].append(endpoint)
                except KeyError:
                    results[category][endpoint.sub_category or "@TopLevelEndpoint"] = [endpoint]
        return results

    def __init__(self,
                 link: str,  # The base link
                 endpoints: typing.Optional[typing.List[Endpoint]] = None,  # Endpoints for the endpoints explorer
                 #  localization: typing.Union[Localization, typing.Type[Localization]] = EnglishLocalization,  # The UI language
                 options: typing.Optional[HTTPOptions] = None,  # Options for the app
                 **kwargs):
        super().__init__(**kwargs)
        self.link = url.urlparse(link)

        # self.localization = localization
        self.endpoints = endpoints or []
        self.base_endpoints = self.endpoints.copy()

        self.options = options or HTTPOptionsScreen.loads("http", HTTPOptions)
        self.localization = language_to_localization(self.options.language)

        self.set_bindings(define_bindings(self.localization))

    @work(exclusive=True)
    def update_endpoints(self, wait: int = 0):
        """The worker thread which updates the endpoints list"""
        worker = get_current_worker()
        time.sleep(wait)
        while True:
            # print("(worker) Checking out the new endpoints")
            try:
                final_path = url.urlunparse((
                    # Defaulting to the registered base URL
                    self.link.scheme,
                    self.link.netloc,
                    "/@nasse/endpoints",
                    "",
                    "",
                    ""
                ))
                response = requests.get(final_path, timeout=60)
                endpoints = []
                for endpoint in response.json()["data"]["endpoints"]:
                    # print(endpoint)
                    endpoint.pop("handler", None)
                    endpoints.append(Endpoint(**endpoint))

            except Exception as err:
                # print("(worker) Opps an error occured while checking for new endpoints", err)
                endpoints = []

            if not worker.is_cancelled:
                self.call_from_thread(self.load_endpoints, endpoints)

            # print(f"(worker) Waiting {self.options.endpoints_update} seconds before checking again for new endpoints")

            # This is a big approximate way of doing this but it's fine because
            # this is not very important
            for _ in range(self.options.endpoints_update):
                if worker.is_cancelled:
                    break
                time.sleep(1)

            if worker.is_cancelled:
                # print("(worker) The endpoints checking got cancelled here")
                break

    def load_endpoints(self, endpoints: typing.List[Endpoint]):
        """Loads the endpoints to the endpoints list"""
        self.endpoints = [*self.base_endpoints, *endpoints]
        # print(f"Found {len(self.endpoints)} endpoints")

    # pylint: disable=pointless-string-statement
    """
    ╭─────────────────────────────── Composers ────────────────────────────────╮
    │ compose                                                                  │
    │ compose_request_view                                                     │
    │ compose_result_start                                                     │
    │ compose_result_error                                                     │
    │ compose_result_loading                                                   │
    │ compose_result_response                                                  │
    │ compose_explorer                                                         │
    ╰──────────────────────────────────────────────────────────────────────────╯
    """

    def compose(self):
        """
        Draws the screen

        Area
        ----
        ╭────────────────────────────────────────────────╮
        │▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒│
        │▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒│
        │▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒│
        │▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒│
        │▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒│
        │▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒│
        │▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒│
        │▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒│
        │▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒│
        ╰────────────────────────────────────────────────╯
        """
        # The header icon will be changed inside the `on_mount` event of this class
        yield Header(show_clock=True)

        # Coverage: Screen
        with Horizontal(id="screen"):
            with Container(id="history"):
                # History
                # Coverage: Left sidebar
                yield StickyHeader(self.localization.tui_history)
                with VerticalScroll(id="history-requests"):
                    # Requests History
                    # This displays a list of already made
                    # requests, whether it be successful or
                    # erroneous ones, in chronological orders.
                    for response in self.history:
                        yield HistoryResponse(response)
                # This displays a graph with the evolution of the time
                # taken for each request
                yield series.TimeSeries([], localization=self.localization, id="history-ping")

            with Container(id="main"):
                # Main Page
                # Coverage: Center of the screen
                yield StickyHeader(self.localization.tui_request, id="request-title")

                # There are actually two views, one for the actual request customisation
                # and one for the request response.
                # They both can be expanded as the user needs by focusing on them.

                # The request customisation
                with View(id="request", on_click=self.on_request_view_clicked):
                    # The actual content is explained below in `compose_request_view`
                    yield from self.compose_request_view()

                # The request response
                with View(id="result", on_click=self.on_result_view_clicked):
                    # Request Result
                    yield StickyHeader(self.localization.tui_result)
                    # This content will be changed by the `compose_result_*` functions
                    # Refer to those functions for further details.
                    yield from self.compose_result_start()

            with Container(id="explorer"):
                # Endpoints explorer
                # Coverage: Right sidebar

                # The endpoints explorer is a feature which lets the user
                # use existing endpoints information to guide them through
                # their use.
                # It prefills the sendable values names, adds a description
                # for items which supports them and allows for a greater
                # understanding of the whole server pathspace.
                yield StickyHeader(self.localization.tui_explorer)

                # This overwrites all of the changes made
                yield Button(self.localization.tui_reset, id="explorer-reset")

                # The actual explorer
                # with VerticalScroll(id="endpoints-explorer"):
                with VerticalScroll(id="tree-view"):
                    yield from self.compose_explorer()

        # Add a footer, which automatically displays the different available bindings
        yield Footer()

    def compose_request_view(self):
        """
        Creates the request customisation view

        Area
        ----
        ╭────────┬──────────────────────────────┬────────╮
        │        |▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒|        │
        │        |▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒|        │
        │        |▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒|        │
        │        |▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒|        │
        │        |▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒|        │
        │        |▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒|        │
        │        ├──────────────────────────────┤        │
        │        |                              |        │
        │        |                              |        │
        ╰────────┴──────────────────────────────┴────────╯

        Note: Comments marked with a `(*)` have method variants, which shows that they might change with the method selected by the user.
        """

        # Caching the selected method
        method = self.method

        if self.endpoint:
            # We can actually display the name and description for defined endpoints
            description = self.endpoint.description.get(method, self.endpoint.description.get("*", self.localization.no_description))
            # We are forced to use this syntax since the context manager doesn't seem to work well
            # with the `mount` and `mount_all` functions
            yield Container(
                Label(self.endpoint.name, id="request-information-title"),  # A box with the endpoint name
                Markdown(description.strip(), id="request-information-description"),  # A blockquote with the endpoint description (*)
                id="request-information"
            )

        if not self.endpoint or not self.endpoint.methods or "*" in self.endpoint.methods:
            # All of the methods might be available
            methods = typing.get_args(Types.Method.Standard)
        else:
            # The endpoint only supports these methods
            methods = self.endpoint.methods

        yield Horizontal(Select([(method, method) for method in methods],
                                allow_blank=False,
                                value=method if method != "*" else "GET",
                                id="request-method"),
                         Input(self.endpoint.path if self.endpoint else "/",
                               placeholder=self.localization.tui_path,
                               suggester=PathSuggestion(self),
                               id="request-path"),
                         id="request-path-container")

        yield Container(UserSentForm(self.localization.parameters, inputs=get_method_variant(method, self.endpoint.parameters)
                                     if self.endpoint else None, multiple=True, id="request-parameters", localization=self.localization),
                        UserSentForm(self.localization.headers, inputs=get_method_variant(method, self.endpoint.headers)
                                     if self.endpoint else None, id="request-headers", localization=self.localization),
                        UserSentForm(self.localization.cookies, inputs=get_method_variant(method, self.endpoint.cookies)
                                     if self.endpoint else None, id="request-cookies", localization=self.localization),
                        id="request-user-sent")

        yield SectionTitle(self.localization.tui_file)
        yield Container(Container(id="request-files-container", classes="files-container"),
                        Button(self.localization.tui_add_file, classes="add-file-button"),
                        classes="files-input-container")

        yield SectionTitle(self.localization.tui_data)
        yield Container(Button(self.localization.tui_add_data_file, id="request-data-button"), id="request-data-container")

    def compose_result_response(self, result: requests.Response):
        """
        Creates the response view for a successful request

        Area
        ----
        ╭────────┬──────────────────────────────┬────────╮
        │        |                              |        │
        │        |                              |        │
        │        |                              |        │
        │        |                              |        │
        │        |                              |        │
        │        |                              |        │
        │        ├──────────────────────────────┤        │
        │        |▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒|        │
        │        |▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒|        │
        ╰────────┴──────────────────────────────┴────────╯
        """
        yield Label(f"{result.request.method} {result.url}", id="result-title")
        yield Label(f"[bold]{result.status_code}[/bold] {result.reason} [grey]in {series.transform_time(result.elapsed.total_seconds() * 1000)}[/grey]", id="result-subtitle")

        if result.headers:
            yield SectionTitle(self.localization.headers)
            yield Label('\n'.join(f'{name}: {file}' for name, file in result.headers.items()))

        if result.cookies:
            yield SectionTitle(self.localization.cookies)
            yield Label('\n'.join(f'{name}: {file}' for name, file in result.cookies.items()))

        content = None
        try:
            content = result.json()
        except Exception:
            pass

        if not content:
            try:
                content = result.text
            except Exception:
                pass

        if not content:
            try:
                content = result.content
            except Exception:
                content = f"[red](ERROR)[/red] {self.localization.tui_no_content}"

        yield SectionTitle(self.localization.tui_content)
        yield Label(f"({sys.getsizeof(content)} bytes)", id="result-size")
        yield Pretty(content, id="result-content")

    def compose_result_start(self):
        """
        Creates the response view when no request is selected.

        Area
        ----
        ╭────────┬──────────────────────────────┬────────╮
        │        |                              |        │
        │        |                              |        │
        │        |                              |        │
        │        |                              |        │
        │        |                              |        │
        │        |                              |        │
        │        ├──────────────────────────────┤        │
        │        |▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒|        │
        │        |▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒|        │
        ╰────────┴──────────────────────────────┴────────╯
        """
        # Prompts the user to make a request, because it is highly likely that no request has been made
        yield Label(self.localization.tui_start_prompt, id="empty-result-label")

    def compose_result_loading(self, result: Loading):
        """
        Creates the response view when a request is being made

        Area
        ----
        ╭────────┬──────────────────────────────┬────────╮
        │        |                              |        │
        │        |                              |        │
        │        |                              |        │
        │        |                              |        │
        │        |                              |        │
        │        |                              |        │
        │        ├──────────────────────────────┤        │
        │        |▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒|        │
        │        |▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒|        │
        ╰────────┴──────────────────────────────┴────────╯
        """
        yield Container(
            Horizontal(
                Label("🧑‍💻"),  # the user
                LoadingIndicator(),  # is making a request
                Label("🌐"),  # to the server
                id="loading-container"
            ),
            Label(self.localization.tui_contacting.format(url=url.urlparse(result.url).netloc)),  # the server hostname
            id="loading-view"
        )

    def compose_result_error(self, result: Error):
        """
        Creates the response view for an error

        Area
        ----
        ╭────────┬──────────────────────────────┬────────╮
        │        |                              |        │
        │        |                              |        │
        │        |                              |        │
        │        |                              |        │
        │        |                              |        │
        │        |                              |        │
        │        ├──────────────────────────────┤        │
        │        |▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒|        │
        │        |▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒|        │
        ╰────────┴──────────────────────────────┴────────╯
        """
        yield Label(f"{result.method} {result.url}", id="result-title")
        yield Label(f"[bold]EXCEPTION[/bold] {result.exception.__class__.__name__}", id="result-subtitle")

        # Showing the sent user values
        if result.params:
            yield SectionTitle(self.localization.parameters)
            yield Label('\n'.join(f'{name}: {file}' for name, file in result.params.items()))

        if result.headers:
            yield SectionTitle(self.localization.headers)
            yield Label('\n'.join(f'{name}: {file}' for name, file in result.headers.items()))

        if result.cookies:
            yield SectionTitle(self.localization.cookies)
            yield Label('\n'.join(f'{name}: {file}' for name, file in result.cookies.items()))

        if result.files:
            yield SectionTitle(self.localization.tui_files)
            yield Label('\n'.join(f'{name}: {file}' for name, file in result.files))

        # Showing the options at the time of the request, because this might
        # the root of the issue (timeouts or SSL issues for example)
        yield SectionTitle(self.localization.tui_options)
        yield Label(f"{self.localization.tui_timeout}: {result.timeout} sec.")
        yield Label(f"{self.localization.tui_allow_redirects}: {result.allow_redirects}")
        if result.proxies:
            yield Label(f"{self.localization.tui_proxies}: {', '.join(f'{prot}: {proxy}' for prot, proxy in result.proxies.items())}")
        yield Label(f"{self.localization.tui_verify_request}: {result.verify}")
        if result.cert:
            yield Label(f"{self.localization.tui_certificate_files}: {', '.join(result.cert)}")

        # Showing a full traceback of the incident to easily debug
        # (this might take a little while to render)
        yield SectionTitle(self.localization.tui_error)
        yield Static(Traceback.from_exception(result.exception.__class__, result.exception, traceback=result.exception.__traceback__), classes="result-error-container")

    def reload_endpoint(self):
        """Reloads the request view"""
        request_view = self.query_one("#request", View)
        request_view.remove_children()
        request_view.mount_all(self.compose_request_view())
        # Making sure the request got refreshed
        request_view.refresh(layout=True)

    def compose_explorer(self):
        """
        Creates the explorer content

        Area
        ----
        ╭────────┬──────────────────────────────┬────────╮
        │        |                              |▒▒▒▒▒▒▒▒│
        │        |                              |▒▒▒▒▒▒▒▒│
        │        |                              |▒▒▒▒▒▒▒▒│
        │        |                              |▒▒▒▒▒▒▒▒│
        │        |                              |▒▒▒▒▒▒▒▒│
        │        |                              |▒▒▒▒▒▒▒▒│
        │        ├──────────────────────────────┤▒▒▒▒▒▒▒▒│
        │        |                              |▒▒▒▒▒▒▒▒│
        │        |                              |▒▒▒▒▒▒▒▒│
        ╰────────┴──────────────────────────────┴────────╯
        """
        for category, sub_categories in self.categories.items():
            # For each category, we are creating a tree, to
            # make a multi-rooted final tree.
            tree: Tree[Endpoint] = Tree(category, classes="explorer-category")
            for sub_category, endpoints in sub_categories.items():
                # An endpoint could have a category without having any sub category
                # This would cause an issue if the developer decides to use
                # `@TopLevelEndpoint` as their sub category name, which is very unlikely
                if sub_category == "@TopLevelEndpoint":
                    sub_tree = tree.root
                else:
                    sub_tree = tree.root.add(sub_category)

                # For each endpoint, add a leaf, either to the tree root, or subtree
                for endpoint in endpoints:
                    sub_tree.add_leaf(endpoint.name, endpoint)
            yield tree

    def reload_explorer(self):
        """Reloads the explorer view"""
        explorer_view = self.query_one("#tree-view", VerticalScroll)
        explorer_view.remove_children()
        explorer_view.mount_all(self.compose_explorer())
        # Making sure the explorer got refreshed
        explorer_view.refresh(layout=True)

    # pylint: disable=pointless-string-statement
    """
    ╭────────────────────────── Event Handlers ────────────────────────────────╮
    │ on_mount                                                                 │
    │ on_select_changed                                                        │
    │ on_view_clicked                                                          │
    │ on_request_view_clicked                                                  │
    │ on_result_view_clicked                                                   │
    │ on_button_pressed                                                        │
    │ on_tree_node_selected                                                    │
    │ add_file                                                                 │
    │ add_data_file                                                            │
    │ delete_file                                                              │
    │ delete_data_file                                                         │
    ╰──────────────────────────────────────────────────────────────────────────╯
    """

    def on_mount(self):
        """When mounted"""
        # Yea it's a pain to change the Header Icon
        self.query_one(Header).query_one(_header.HeaderIcon).icon = "🍡"

        # Start to update the endpoints list
        self.update_endpoints()

        # I thought I would have a problem with the mounting flow
        # because the `on_mount` method would be called before
        # drawing the DOM, thus creating a problem within `watch_endpoints`
        # but it doesn't seem to be the case ?
        # self.update_endpoints(wait=3)

    def on_select_changed(self, event: Select.Changed):
        """When a selected element is changed"""
        if event.select.id == "request-method":
            self._method = event.select.value or "*"
            self.reload_endpoint()

    def on_view_clicked(self, minimizing: str, maximizing: str):
        """When a view is clicked"""
        minimizing_element = self.query_one(f"#{minimizing}", VerticalScroll)
        maximizing_element = self.query_one(f"#{maximizing}", VerticalScroll)

        if maximizing_element.has_class("mini"):
            for element in (minimizing_element, maximizing_element):
                element.remove_class("mini")
                element.remove_class("maxi")
        else:
            maximizing_element.add_class("maxi")
            minimizing_element.add_class("mini")
            minimizing_element.remove_class("maxi")  # it normally shouldn't have this

    def on_request_view_clicked(self, event: Click):
        """When the request view is clicked"""
        self.on_view_clicked(minimizing="result", maximizing="request")

    def on_result_view_clicked(self, event: Click):
        """When the request view is clicked"""
        self.on_view_clicked(minimizing="request", maximizing="result")

    def on_button_pressed(self, event: Button.Pressed) -> None:
        """When a button is pressed"""
        # When it's an element in the request history
        if isinstance(event.button, HistoryResponse):
            # Just show the request in the `Result` panel
            self.result = event.button.response
        # When we are adding a file to the request
        if event.button.has_class("add-file-button"):
            # Show the file browser to add the file to the list
            self.push_screen(FileBrowser(localization=self.localization), self.add_file)
        # When we are adding binary data to the request
        if event.button.id == "request-data-button":
            # Show the file browser to add get the file where the data will be fetched
            self.push_screen(FileBrowser(localization=self.localization), self.add_data_file)
        # If we are resetting the endpoint
        if event.button.id == "explorer-reset":
            # Clearing the request view
            self.endpoint = None
            self.reload_endpoint()

    def on_tree_node_selected(self, event: Tree.NodeSelected):
        """When a node is selected in a tree"""
        # Should be in the endpoints explorer tree
        if isinstance(event.node.data, Endpoint):
            self.endpoint = event.node.data
            self.reload_endpoint()

    def add_file(self, file: typing.Optional[pathlib.Path] = None):
        """Adds a file to the request"""
        if file:
            self.query_one("#request-files-container", Container).mount(FileInput(file, on_delete=self.delete_file))

    def add_data_file(self, file: typing.Optional[pathlib.Path] = None):
        """Adds the data file to the request"""
        if file:
            container = self.query_one("#request-data-container", Container)
            container.remove_children()
            container.mount(FileInput(file,
                                      on_delete=self.delete_data_file,
                                      prompt_name=False,  # data is not named
                                      id="request-data-file"))

    def delete_file(self, file_input: FileInput, file: pathlib.Path):
        """When a file is removed from the files list"""
        file_input.remove()

    def delete_data_file(self, file_input: FileInput, file: pathlib.Path):
        """When the data file is removed"""
        container = self.query_one("#request-data-container", Container)
        container.remove_children()
        container.mount(Button(self.localization.tui_add_data_file, id="request-data-button"))

    def on_url_change(self, link: str):
        """When the base URL got changed in the settings"""
        self.link = url.urlparse(str(link))
        self.update_endpoints()

    # pylint: disable=pointless-string-statement
    """
    ╭────────────────────────── Binding actions ───────────────────────────────╮
    │ action_open_options                                                      │
    │ replace_options                                                          │
    │ action_request_quit                                                      │
    │ action_toggle_history                                                    │
    │ action_toggle_results                                                    │
    │ action_toggle_explorer                                                   │
    │ action_submit                                                            │
    │ request_worker                                                           │
    │ add_result                                                               │
    ╰──────────────────────────────────────────────────────────────────────────╯
    """

    def action_open_options(self):
        """When the user wants to see the options screen"""
        self.push_screen(HTTPOptionsScreen(base_url=url.urlunparse(self.link),
                                           on_url_change=self.on_url_change,
                                           options=self.options,
                                           id="options-screen",
                                           localization=self.localization),
                         self.replace_options)

    def replace_options(self, options: HTTPOptions):
        """To replace the current options"""
        self.options = options
        HTTPOptionsScreen.dumps("http", options)
        self.localization = language_to_localization(self.options.language)

    def action_request_quit(self) -> None:
        """Action to display the quit dialog."""
        self.push_screen(QuitScreen(localization=self.localization))

    def action_toggle_history(self):
        """Called when the user fires the `toggle_history` action"""
        self.toggle_history = not self.toggle_history

    def action_toggle_results(self):
        """Called when the user fires the `toggle_results` action"""
        self.toggle_results = not self.toggle_results

    def action_toggle_explorer(self):
        """Called when the user fires the `toggle_explorer` action"""
        self.toggle_explorer = not self.toggle_explorer

    def action_submit(self):
        """Called when the user submitted the request"""
        method = self.query_one("#request-method", Select).value or "*"
        path = url.urlparse(self.query_one("#request-path", Input).value)
        final_path = url.urlunparse((
            # Defaulting to the registered base URL
            path.scheme or self.link.scheme,
            path.netloc or self.link.netloc,
            path.path,
            path.params,
            path.query,
            path.fragment
        ))

        params = self.query_one("#request-parameters", UserSentForm).values

        # Setting a custom user agent by default
        headers = {
            "User-Agent": f"nasse-http/{__info__.__version__}"
        }
        # Which can be overwritten here
        headers.update(self.query_one("#request-headers", UserSentForm).values)

        cookies = self.query_one("#request-cookies", UserSentForm).values

        self.result = Loading(url=final_path)

        files = []
        for file_input in self.query_one("#request-files-container", Container).query(FileInput):
            files.append((file_input.input_name or "file", (file_input.file.name, file_input.file.open("rb"))))

        try:
            data = self.query_one("#request-data-file", FileInput).file.read_bytes()
        except Exception:
            data = None

        self.request_worker(method, final_path, params, headers, cookies, files, data)

    @work(exclusive=True)
    def request_worker(self,
                       method: str,
                       final_path: str,
                       params: typing.Dict[str, typing.List[str]],
                       headers: typing.Dict[str, str],
                       cookies: typing.Dict[str, str],
                       files: typing.List[typing.Tuple[str, typing.Tuple[str, typing.BinaryIO]]],
                       data: typing.Optional[bytes] = None):
        """The worker thread which actually makes the request"""
        worker = get_current_worker()
        try:
            if not self.options.cert:
                cert = None
            elif len(self.options.cert) == 1:
                cert = self.options.cert[0]
            else:
                # it can't exceed two files
                cert = (self.options.cert[0], self.options.cert[1])
            response = requests.request(method,
                                        final_path,
                                        params=params,
                                        headers=headers,
                                        cookies=cookies,
                                        files=files,
                                        data=data,
                                        timeout=self.options.timeout,
                                        allow_redirects=self.options.allow_redirects,
                                        verify=self.options.verify,
                                        proxies=self.options.proxies,
                                        cert=cert)
        except Exception as exc:
            response = Error(exception=exc,
                             method=method,
                             url=final_path,
                             params=params,
                             headers=headers,
                             cookies=cookies,
                             files=[(name, element[0]) for name, element in files],
                             data=data,
                             timeout=self.options.timeout,
                             allow_redirects=self.options.allow_redirects,
                             verify=self.options.verify,
                             proxies=self.options.proxies,
                             cert=self.options.cert)

        if not worker.is_cancelled:
            self.call_from_thread(self.add_result, response)
            # Opening up the result panel if we got a result
            self.toggle_results = True

    def add_result(self, response: typing.Union[requests.Response, Error]):
        """Adds the given response to the results"""
        self.history = [*self.history, response]
        self.result = response

    # pylint: disable=pointless-string-statement
    """
    ╭──────────────────────── Reactive handlers ────────────────────────────╮
    │ watch_localization                                                    │
    │ watch_endpoints                                                       │
    │ watch_toggle_history                                                  │
    │ watch_toggle_results                                                  │
    │ watch_toggle_explorer                                                 │
    │ watch_history                                                         │
    │ watch_result                                                          │
    ╰───────────────────────────────────────────────────────────────────────╯
    """

    def watch_localization(self, localization: typing.Type[Localization]) -> None:
        """Called when `localization` is modified"""
        # self.refresh(repaint=True, layout=True)
        self.app.exit()
        self.app.run()

    def watch_endpoints(self, endpoints: typing.List[Endpoint]) -> None:
        """Called when `endpoints` is modified"""
        try:
            self.reload_explorer()
        except Exception:
            # print("Couldn't reload the explorer")
            pass

    def watch_toggle_history(self, toggle_history: bool) -> None:
        """Called when `toggle_history` is modified"""
        self.query_one("#history", Container).set_class(not toggle_history, "unload")

    def watch_toggle_results(self, toggle_results: bool) -> None:
        """Called when `toggle_results` is modified"""
        self.query_one("#result", VerticalScroll).set_class(not toggle_results, "unload")
        self.query_one("#request", VerticalScroll).set_class(not toggle_results, "full")

    def watch_toggle_explorer(self, toggle_explorer: bool) -> None:
        """Called when `toggle_explorer` is modified"""
        self.query_one("#explorer", Container).set_class(not toggle_explorer, "unload")

    def watch_history(self, history: typing.List[requests.Response]) -> None:
        """Called when `history` is modified"""
        # Avoiding to store too much things in the request history
        # because each one of them is rendered in the history panel
        self.history = history[-(self.options.history_limit - 1):]
        try:
            self.query_one(series.TimeSeries).series = [resp.elapsed.total_seconds() * 1000  # measured in ms
                                                        for resp in history if isinstance(resp, requests.Response)]
            history_requests = self.query_one("#history-requests", VerticalScroll)
            history_requests.remove_children()
            history_requests.mount_all([HistoryResponse(element) for element in reversed(history)])
        except Exception:
            pass

    def watch_result(self, result: typing.Optional[requests.Response]):
        """Called when `result` is modified"""
        container = self.query_one("#result", VerticalScroll)
        container.remove_children()
        container.mount(StickyHeader(self.localization.tui_result))

        if result is None:
            return container.mount_all(self.compose_result_start())
        if isinstance(result, Loading):
            return container.mount_all(self.compose_result_loading(result))
        if isinstance(result, Error):
            return container.mount_all(self.compose_result_error(result))

        return container.mount_all(self.compose_result_response(result))

# Made this path suggestion thing, which searches
# its suggestions in the request history but doesn't seem to
# work well and I don't really know why


class PathSuggestion(Suggester):
    """Gives HTTP path suggestions"""

    def __init__(self, app: HTTP) -> None:
        super().__init__(use_cache=False, case_sensitive=False)
        self.app = app

    async def get_suggestion(self, value: str) -> typing.Optional[str]:
        """Gets a completion from the given possibilities.

        Args:
            value: The current value.

        Returns:
            A valid completion suggestion or `None`.
        """
        for element in self.app.history:
            path = url.urlparse(element.url).path
            if value.startswith(path):
                return path
        return None


if __name__ == "__main__":
    # HTTP("https://google.com").run()
    # HTTP("https://eosqyydyun9tw26.m.pipedream.net").run()
    # HTTP("http://httpbin.org/get", endpoints=[
    #     Endpoint(name="GET request", category="Method Requests", sub_category="GET", methods="GET",
    #              description="This is a GET request", headers=[UserSent("X-NASSE-TEST", description="This is a test")], path="/get"),
    #     Endpoint(name="POST request", category="Method Requests", sub_category="POST", methods="POST",
    #              description="This is a POST request", headers=[UserSent("X-NASSE-TEST", description="This is a test")], parameters=UserSent("hello", description="world"), path="/post"),
    #     Endpoint(name="Multiple request", category="Method Requests", methods="*",
    #              description={"GET": "This is a multiple methods request", "POST": "This is really cool", "*": "Yup as expected"}, headers=[UserSent("X-NASSE-TEST", description="This is a test")], parameters={"POST": UserSent("hello", description="world")}, path="/post"),
    # ] * 20).run()
    HTTP("").run()



"""
A utility to retrieve the IP address of the client
"""
import flask


def get_ip():
    """
    Retrieves the client IP address from the current request
    """
    # if the server uses a proxy
    if "HTTP_X_FORWARDED_FOR" in flask.request.environ:
        x_forwarded_for = str(flask.request.environ['HTTP_X_FORWARDED_FOR']).split(',')[0]
        try:
            if x_forwarded_for.replace('.', '').isdigit():
                return x_forwarded_for
            else:
                return flask.request.remote_addr
        except Exception:
            return flask.request.remote_addr
    else:
        return flask.request.remote_addr



from nasse.exceptions.base import NasseException


class ClientError(NasseException):
    """When there is a client side error"""
    STATUS_CODE = 400
    MESSAGE = "Something is missing from the request"
    EXCEPTION_NAME = "CLIENT_ERROR"

class InvalidType(ClientError):
    """When the given parameter is of an invalid type"""
    MESSAGE = "The given value is of an invalid type"
    EXCEPTION_NAME = "INVALID_TYPE"

    def __init__(self, *args: object, name: str = "") -> None:
        message = "The given value `{name}` could not be casted to a valid type".format(name=name)
        super().__init__(message=message, *args)

class MissingValue(ClientError):
    """When a value is missing from the request"""
    MESSAGE = "A value is missing from your request"
    EXCEPTION_NAME = "MISSING_VALUE"

    def __init__(self, *args: object, name: str = "", missing_type: str = "value") -> None:
        message = "`{name}` is a required request {type}".format(name=name,
                                                                 type=str(missing_type))
        super().__init__(message=message, *args)


class MissingParam(MissingValue):
    """When a parameter is missing"""
    MESSAGE = "A parameter is missing from your request"
    EXCEPTION_NAME = "MISSING_PARAM"


class MissingDynamic(MissingValue):
    """When a dynamic routing value is missing"""
    MESSAGE = "A dynamic routing value is missing from your request"
    EXCEPTION_NAME = "MISSING_DYNAMIC"


class MissingHeader(MissingValue):
    """When a header is missing"""
    MESSAGE = "A header is missing from your request"
    EXCEPTION_NAME = "MISSING_HEADER"


class MissingCookie(MissingValue):
    """When a cookie is missing"""
    MESSAGE = "A cookie is missing from your request"
    EXCEPTION_NAME = "MISSING_COOKIE"


class MissingContext(NasseException):
    """Server side error when a value only accessible in a Nasse context is accessed outside of oneƒ"""
    MESSAGE = "You are not actively in a Nasse context"
    EXCEPTION_NAME = "MISSING_CONTEXT"
    STATUS_CODE = 500


class MissingEndpoint(MissingContext):
    """When an endpoint is missing"""
    # ??



from textual import events, keys
from textual.widgets import Input as DefaultInput


class Input(DefaultInput):
    """A (custom) text input widget."""

    def on_key(self, event: events.Key) -> None:
        """Checks for `enter` or `esc` to blur the input"""
        if event.key in (keys.Keys.Enter, keys.Keys.Escape):
            try:
                self.screen.set_focus(None)
            except Exception:
                self.blur()



"""
A utility to parse boolean values
"""
import typing

from nasse import utils


def to_bool(value: typing.Any, default: bool = False):
    """
    Converts any value to a boolean

    Example
    -------
    >>> from nasse.utils.boolean import to_bool
    >>> to_bool("true")
    True
    >>> to_bool(" 1 ")
    True
    >>> to_bool(1)
    True
    >>> to_bool(" yes!")
    False
    >>> to_bool(" yes!", default=True)
    True
    >>> to_bool(" no ", default=True)
    False


    Parameters
    ----------
        value: Any
            The value to convert into a boolean
        default: bool
            The default value
    """
    if utils.sanitize.remove_spaces(value).lower() in ({"true", "1", "yes"} if not default else {"false", "0", "no"}):
        return not default
    return default



"""
Generates JavaScript examples
"""

import typing
import urllib.parse

from nasse import models, utils


def create_javascript_example_for_method(endpoint: models.Endpoint, method: str) -> str:
    """
    Creates a JavaScript example for the given method

    Parameters
    ----------
    endpoint: models.Endpoint
        The endpoint to create the example from
    method: str
        The method to get the example for

    Returns
    -------
    str
        A JavaScript example on how to use the endpoint for the given method
    """
    params = [param for param in models.get_method_variant(method, endpoint.parameters) if param.required]
    headers = {header.name: (header.description or header.name)
               for header in models.get_method_variant(method, endpoint.headers)
               if header.required}

    include_cookies = len(models.get_method_variant(method, endpoint.cookies)) > 0

    url = '"{path}"'.format(path=endpoint.path)

    if len(params) > 0:
        url = """`{path}?{params}`""".format(
            path=endpoint.path,
            params="&".join([
                '{escaped}=${{encodeURIComponent("{name}")}}'.format(
                    escaped=urllib.parse.quote(param.name),
                    name=param.name)
                for param in params]))

    headers_render = ""
    if len(headers) > 0:
        headers_render = ",\n    headers: {}".format(utils.json.encoder.encode(headers).replace("\n", "\n    "))

    return '''fetch({url}, {{
    method: "{method}"{headers}{cookies}
}})
.then((response) => {{response.json()}})
.then((response) => {{
    if (response.success) {{
        console.info("Successfully requested for {path}")
        console.log(response.data)
    }} else {{
        console.error("An error occured while requesting for {path}, error: " + response.error)
    }}
}})'''.format(url=url, method=method, headers=headers_render, cookies=",\n    credentials: 'include'" if include_cookies else "", path=endpoint.path)


def create_javascript_example(endpoint: models.Endpoint) -> typing.Dict[str, str]:
    """
    Creates JavaScript examples for the given endpoint

    Parameters
    ----------
    endpoint: models.Endpoint
        The endpoint to get the example for

    Returns
    -------
    """
    results = {}
    for method in endpoint.methods:
        results[method] = create_javascript_example_for_method(endpoint=endpoint,
                                                               method=method)
    return results



"""
The Nasse JSON Encoder

Most of the code here comes from the builtin `json` library
but slight changes got applied.

Copyright © 2001-2023 Python Software Foundation.

License
-------

1. This LICENSE AGREEMENT is between the Python Software Foundation
("PSF"), and the Individual or Organization ("Licensee") accessing and
otherwise using this software ("Python") in source or binary form and
its associated documentation.

2. Subject to the terms and conditions of this License Agreement, PSF hereby
grants Licensee a nonexclusive, royalty-free, world-wide license to reproduce,
analyze, test, perform and/or display publicly, prepare derivative works,
distribute, and otherwise use Python alone or in any derivative version,
provided, however, that PSF's License Agreement and PSF's notice of copyright,
i.e., "Copyright (c) 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009, 2010,
2011, 2012, 2013, 2014, 2015, 2016, 2017, 2018, 2019, 2020, 2021, 2022, 2023 Python Software Foundation;
All Rights Reserved" are retained in Python alone or in any derivative version
prepared by Licensee.

3. In the event Licensee prepares a derivative work that is based on
or incorporates Python or any part thereof, and wants to make
the derivative work available to others as provided herein, then
Licensee hereby agrees to include in any such work a brief summary of
the changes made to Python.

4. PSF is making Python available to Licensee on an "AS IS"
basis.  PSF MAKES NO REPRESENTATIONS OR WARRANTIES, EXPRESS OR
IMPLIED.  BY WAY OF EXAMPLE, BUT NOT LIMITATION, PSF MAKES NO AND
DISCLAIMS ANY REPRESENTATION OR WARRANTY OF MERCHANTABILITY OR FITNESS
FOR ANY PARTICULAR PURPOSE OR THAT THE USE OF PYTHON WILL NOT
INFRINGE ANY THIRD PARTY RIGHTS.

5. PSF SHALL NOT BE LIABLE TO LICENSEE OR ANY OTHER USERS OF PYTHON
FOR ANY INCIDENTAL, SPECIAL, OR CONSEQUENTIAL DAMAGES OR LOSS AS
A RESULT OF MODIFYING, DISTRIBUTING, OR OTHERWISE USING PYTHON,
OR ANY DERIVATIVE THEREOF, EVEN IF ADVISED OF THE POSSIBILITY THEREOF.

6. This License Agreement will automatically terminate upon a material
breach of its terms and conditions.

7. Nothing in this License Agreement shall be deemed to create any
relationship of agency, partnership, or joint venture between PSF and
Licensee.  This License Agreement does not grant permission to use PSF
trademarks or trade name in a trademark sense to endorse or promote
products or services of Licensee, or any third party.

8. By copying, installing or otherwise using Python, Licensee
agrees to be bound by the terms and conditions of this License
Agreement.
"""

import base64
import io
import json
import json.encoder
import typing
import dataclasses

from nasse import utils

PYTHON_DEFAULT_DECODER = json.JSONEncoder().default


class NasseJSONEncoder(json.JSONEncoder):
    """A custom JSON encoder"""

    def _make_iterencode(self, markers, _default, _encoder, _indent, _floatstr,
                         _key_separator, _item_separator, _sort_keys, _skipkeys, _one_shot,
                         # HACK: hand-optimized bytecode; turn globals into locals
                         ValueError=ValueError,
                         dict=dict,
                         float=float,
                         id=id,
                         int=int,
                         isinstance=isinstance,
                         list=list,
                         str=str,
                         tuple=tuple,
                         _intstr=int.__repr__,
                         ):

        if _indent is not None and not isinstance(_indent, str):
            _indent = ' ' * _indent

        def _iterencode_list(lst, _current_indent_level):
            lst = self.default(lst)
            if not lst:
                yield '[]'
                return
            if markers is not None:
                markerid = id(lst)
                if markerid in markers:
                    raise ValueError("Circular reference detected")
                markers[markerid] = lst
            buf = '['
            if _indent is not None:
                _current_indent_level += 1
                newline_indent = '\n' + _indent * _current_indent_level
                separator = _item_separator + newline_indent
                buf += newline_indent
            else:
                newline_indent = None
                separator = _item_separator
            first = True
            for value in lst:
                if first:
                    first = False
                else:
                    buf = separator
                if isinstance(value, str):
                    yield buf + _encoder(value)
                elif value is None:
                    yield buf + 'null'
                elif value is True:
                    yield buf + 'true'
                elif value is False:
                    yield buf + 'false'
                elif isinstance(value, int):
                    # Subclasses of int/float may override __repr__, but we still
                    # want to encode them as integers/floats in JSON. One example
                    # within the standard library is IntEnum.
                    yield buf + _intstr(value)
                elif isinstance(value, float):
                    # see comment above for int
                    yield buf + _floatstr(value)
                else:
                    yield buf
                    if isinstance(value, (list, tuple)):
                        chunks = _iterencode_list(value, _current_indent_level)
                    elif isinstance(value, dict):
                        chunks = _iterencode_dict(value, _current_indent_level)
                    else:
                        chunks = _iterencode(value, _current_indent_level)
                    yield from chunks
            if newline_indent is not None:
                _current_indent_level -= 1
                yield '\n' + _indent * _current_indent_level
            yield ']'
            if markers is not None:
                try:
                    del markers[markerid]
                except Exception:
                    pass

        def _iterencode_dict(dct, _current_indent_level):
            dct = self.default(dct)
            if not dct:
                yield '{}'
                return
            if markers is not None:
                markerid = id(dct)
                if markerid in markers:
                    raise ValueError("Circular reference detected")
                markers[markerid] = dct
            yield '{'
            if _indent is not None:
                _current_indent_level += 1
                newline_indent = '\n' + _indent * _current_indent_level
                item_separator = _item_separator + newline_indent
                yield newline_indent
            else:
                newline_indent = None
                item_separator = _item_separator
            first = True
            if _sort_keys:
                items = sorted(dct.items())
            else:
                items = dct.items()
            for key, value in items:
                if isinstance(key, str):
                    pass
                # JavaScript is weakly typed for these, so it makes sense to
                # also allow them.  Many encoders seem to do something like this.
                elif isinstance(key, float):
                    # see comment for int/float in _make_iterencode
                    key = _floatstr(key)
                elif key is True:
                    key = 'true'
                elif key is False:
                    key = 'false'
                elif key is None:
                    key = 'null'
                elif isinstance(key, int):
                    # see comment for int/float in _make_iterencode
                    key = _intstr(key)
                elif _skipkeys:
                    continue
                else:
                    raise TypeError(f'keys must be str, int, float, bool or None, '
                                    f'not {key.__class__.__name__}')
                if first:
                    first = False
                else:
                    yield item_separator
                yield _encoder(key)
                yield _key_separator
                if isinstance(value, str):
                    yield _encoder(value)
                elif value is None:
                    yield 'null'
                elif value is True:
                    yield 'true'
                elif value is False:
                    yield 'false'
                elif isinstance(value, int):
                    # see comment for int/float in _make_iterencode
                    yield _intstr(value)
                elif isinstance(value, float):
                    # see comment for int/float in _make_iterencode
                    yield _floatstr(value)
                else:
                    if isinstance(value, (list, tuple)):
                        chunks = _iterencode_list(value, _current_indent_level)
                    elif isinstance(value, dict):
                        chunks = _iterencode_dict(value, _current_indent_level)
                    else:
                        chunks = _iterencode(value, _current_indent_level)
                    yield from chunks
            if newline_indent is not None:
                _current_indent_level -= 1
                yield '\n' + _indent * _current_indent_level
            yield '}'
            if markers is not None:
                try:
                    del markers[markerid]
                except Exception:
                    pass

        def _iterencode(o, _current_indent_level):
            o = _default(o)
            if isinstance(o, str):
                yield _encoder(o)
            elif o is None:
                yield 'null'
            elif o is True:
                yield 'true'
            elif o is False:
                yield 'false'
            elif isinstance(o, int):
                # see comment for int/float in _make_iterencode
                yield _intstr(o)
            elif isinstance(o, float):
                # see comment for int/float in _make_iterencode
                yield _floatstr(o)
            elif isinstance(o, (list, tuple)):
                yield from _iterencode_list(o, _current_indent_level)
            elif isinstance(o, dict):
                yield from _iterencode_dict(o, _current_indent_level)
            else:
                if markers is not None:
                    markerid = id(o)
                    if markerid in markers:
                        raise ValueError("Circular reference detected")
                    markers[markerid] = o
                raise TypeError('Unsupported type: {cls}'.format(cls=o.__class__.__name__))
        return _iterencode

    def encode_bytes(self, b: bytes) -> str:
        return base64.b64encode(b).decode("utf-8")

    def encode_file(self, f: io.BytesIO):
        position = f.tell()  # storing the current position
        content = f.read()  # read it (place the cursor at the end)
        f.seek(position)  # go back to the original position
        if "b" in f.mode:  # if binary mode
            return self.encode_bytes(content)
        else:
            return str(content)

    def encode_iterable(self, a: typing.Iterable):
        return list(a)

    def iterencode(self, o, _one_shot=False):
        """Encode the given object and yield each string
        representation as available.

        For example::

            for chunk in JSONEncoder().iterencode(bigobject):
                mysocket.write(chunk)

        """
        if self.check_circular:
            markers = {}
        else:
            markers = None
        if self.ensure_ascii:
            _encoder = json.encoder.py_encode_basestring_ascii
        else:
            _encoder = json.encoder.py_encode_basestring

        def floatstr(o, allow_nan=self.allow_nan,
                     _repr=float.__repr__, _inf=json.encoder.INFINITY, _neginf=-json.encoder.INFINITY):
            # Check for specials.  Note that this type of test is processor
            # and/or platform-specific, so do tests which don't depend on the
            # internals.

            if o != o:
                text = 'NaN'
            elif o == _inf:
                text = 'Infinity'
            elif o == _neginf:
                text = '-Infinity'
            else:
                return _repr(o)

            if not allow_nan:
                raise ValueError(
                    "Out of range float values are not JSON compliant: " +
                    repr(o))

            return text

        return self._make_iterencode(
            markers, self.default, _encoder, self.indent, floatstr,
            self.key_separator, self.item_separator, self.sort_keys,
            self.skipkeys, _one_shot)(o, 0)

    def default(self, o: typing.Any) -> typing.Any:
        if dataclasses.is_dataclass(o):
            o = dataclasses.asdict(o)
        if isinstance(o, str):
            return json.encoder.py_encode_basestring(o)
        elif isinstance(o, list):  # some classes inheriting from list might have implemented methods to recognize them as unpackable
            return self.encode_iterable(o)
        elif utils.unpack.is_unpackable(o):
            return dict(o)
        elif isinstance(o, bytes):
            return self.encode_bytes(o)
        elif hasattr(o, "read") and hasattr(o, "tell") and hasattr(o, "seek"):
            return self.encode_file(o)
        elif isinstance(o, typing.Iterable):
            return self.encode_iterable(o)
        try:
            return PYTHON_DEFAULT_DECODER(o)
        except TypeError:
            pass
        utils.logging.logger.debug("Object of type <{type}> will be converted to str while encoding to JSON".format(type=o.__class__.__name__))
        return str(o)


encoder = NasseJSONEncoder(ensure_ascii=False, indent=4)
minified_encoder = NasseJSONEncoder(ensure_ascii=False, separators=(",", ":"))



"""
A string formatter for Nasse

Copyright
---------
Animenosekai
    MIT License
"""
import datetime
import enum
import inspect
import os
import threading
import typing
from string import Formatter

from nasse.__info__ import __version__


class Colors(enum.Enum):
    """A set of ANSI colors"""
    NORMAL = '\033[0m'
    GREY = '\033[90m'
    RED = '\033[91m'
    GREEN = '\033[92m'
    BLUE = '\033[94m'
    CYAN = '\033[96m'
    WHITE = '\033[97m'
    YELLOW = '\033[93m'
    MAGENTA = '\033[95m'


def caller_name(skip: int = 2):
    """
    Get a name of a caller in the format module.class.method

    `skip` specifies how many levels of stack to skip while getting caller
    name. skip=1 means "who calls me", skip=2 "who calls my caller" etc.

    An empty string is returned if skipped levels exceed stack height

    Note: https://stackoverflow.com/a/9812105/11557354

    Parameters
    ----------
    skip: int, default = 2
    """
    stack = inspect.stack()
    start = 0 + skip
    if len(stack) < start + 1:
        return ''
    parentframe = stack[start][0]
    name = []
    module = inspect.getmodule(parentframe)
    if module:
        name.append(module.__name__)
    if 'self' in parentframe.f_locals:
        name.append(parentframe.f_locals['self'].__class__.__name__)
    codename = parentframe.f_code.co_name
    if codename != '<module>':  # top level usually
        name.append(codename)  # function or a method
    del parentframe, stack
    return ".".join(name)


class Unformatted:
    """A dummy object used by the silent formatter"""

    def __init__(self, key):
        self.key = key

    def __format__(self, format_spec):
        return "{{{}{}}}".format(self.key, ":" + format_spec if format_spec else "")


class SilentFormatter(Formatter):
    """
    A formatter which silences the KeyErrors and IndexErrors

    Author
    ------
    CodeManX
        https://stackoverflow.com/a/21754294/11557354
    """

    def get_value(self, key, args, kwargs):
        if isinstance(key, int):
            try:
                return args[key]
            except IndexError:
                return Unformatted(key)
        else:
            try:
                return kwargs[key]
            except KeyError:
                return Unformatted(key)

# pylint: disable=redefined-builtin


def format(string: str, time_format: typing.Union[str, typing.Callable[[datetime.datetime], typing.Any]] = "%Y/%m/%d, %H:%M:%S", config: "NasseConfig" = None, *args, **kwargs):
    """
    Formats the string with the given config

    Parameters
    ----------
    string: str
    time_format: typing.Union[str, typing.Callable[[datetime.datetime], typing.Any]], default = "%Y/%m/%d %H: %M:%S"
    config: NasseConfig, default = None
    level: str, default = None
    """
    formatting = {
        # colors
        "normal": Colors.NORMAL.value,
        "grey": Colors.GREY.value,
        "gray": Colors.GREY.value,
        "red": Colors.RED.value,
        "green": Colors.GREEN.value,
        "blue": Colors.BLUE.value,
        "cyan": Colors.CYAN.value,
        "turquoise": Colors.CYAN.value,
        "white": Colors.WHITE.value,
        "yellow": Colors.YELLOW.value,
        "purple": Colors.MAGENTA.value,
        "pink": Colors.MAGENTA.value,
        "magenta": Colors.MAGENTA.value
    }
    if config:
        formatting.update({
            "name": config.name,
            "app": config.name,
            "id": config.id,
            "host": config.host,
            "port": config.port,
            "debug": config.debug,
            "version": __version__,
            "base_dir": config.base_dir
        })
    time = datetime.datetime.now()
    if "{time}" in string:  # current time
        formatting["time"] = time_format(time) if callable(time_format) else time.strftime(time_format)
    if "{caller}" in string:  # caller function
        formatting["caller"] = caller_name()
    if "{thread}" in string:  # thread id
        formatting["thread"] = threading.get_ident()
    if "{pid}" in string:  # process id
        formatting["pid"] = os.getpid()
    if "{cwd}" in string:  # current working directory
        formatting["cwd"] = os.getcwd()

    # Might optionally add stuff from psutil

    return SilentFormatter().format(string, *args, **{
        **formatting,
        **kwargs
    })



"""
This is the flask default server backend
"""
import werkzeug
from nasse import config, servers


class Flask(servers.ServerBackend):
    """
    The default Flask server backend
    """

    def __init__(self, app: "Nasse", config: config.NasseConfig) -> None:
        self.app = app
        self.config = config or config.NasseConfig()
        self.server = None

    def run(self, *args, **kwargs):
        # kwargs.setdefault("use_reloader", self.config.debug)
        # kwargs.setdefault("use_debugger", self.config.debug)
        kwargs.setdefault("threaded", True)

        # make_server does not support `debug`
        kwargs.pop("debug", None)

        self.server = werkzeug.serving.make_server(
            host=self.config.host,
            port=self.config.port,
            app=self.app.flask,
            *args,
            **kwargs
        )
        self.server.serve_forever()

    def stop(self):
        if self.server is not None:
            self.server.shutdown()



"""
The Gunicorn backend

Note: Should be more suitable for production use
"""

import multiprocessing

from nasse import Nasse, config
from nasse.servers.flask import Flask

import gunicorn
import gunicorn.app.base
import gunicorn.arbiter


class Gunicorn(Flask):
    """
    The Gunicorn backend
    """
    class BaseApp(gunicorn.app.base.BaseApplication):
        """
        The internal Gunicorn base app
        """

        def __init__(self, app: Nasse, config: config.NasseConfig = None, **kwargs):
            self.app = app
            self.config = config or config.NasseConfig()
            self.options = {
                'bind': '{host}:{port}'.format(host=self.config.host, port=self.config.port),
                'workers': 2 * multiprocessing.cpu_count() + 1,
                'capture_output': False,
                'proc_name': self.app.id,
                'preload_app': True,
                'worker_class': "sync",
                'threads': 2 * multiprocessing.cpu_count() + 1,
                'loglevel': 'error',
                # would be painful to wait 30 seconds on each reload
                'graceful_timeout': 5 if self.config.debug else 20,
                'on_exit': self.on_exit,
                'on_starting': self.on_starting,
                **kwargs
            }
            self.options.update(kwargs or {})
            self.application = app.flask
            gunicorn.SERVER_SOFTWARE = self.config.server_header
            gunicorn.SERVER = gunicorn.SERVER_SOFTWARE
            super().__init__()

        def load_config(self):
            config = {key: value for key, value in self.options.items()
                      if key in self.cfg.settings and value is not None}
            for key, value in config.items():
                self.cfg.set(key.lower(), value)

        def load(self):
            return self.app.flask

        def on_starting(self, server):
            self.config.logger.log("Running the server ✨")

        def on_exit(self, server):
            self.config.logger.log("Exiting... 🏮")

    def __init__(self, app: Nasse, config: config.NasseConfig) -> None:
        super().__init__(app, config)
        self._arbiter = None

    def run(self, *args, **kwargs):
        gunicorn_handler = self.BaseApp(self.app, self.config, **kwargs)
        self._arbiter = gunicorn.arbiter.Arbiter(gunicorn_handler)
        self._arbiter.run()

    def stop(self, *args, **kwargs):
        if self._arbiter:
            self._arbiter.stop()



"""
This is where the request context is created and gets sanitized
"""

import typing

import flask
import werkzeug.datastructures

from nasse import config, exceptions, models, utils

_overwritten = {"nasse", "app", "nasse_endpoint",
                "client_ip", "method", "headers", "values", "args", "form", "params", "cookies"}


class Request(object):
    """Represents the current request"""

    def __init__(self, app, endpoint: models.Endpoint, dynamics: dict = None) -> None:
        """
        A request object looking like the flask.Request one, but with the current endpoint in it and verification

        Example
        --------
        >>> from nasse import request
        # and then when processing the request #
        >>> request.nasse_endpoint
        Endpoint(path='/hello')
        >>> request.nasse_endpoint.name
        'Greeting Endpoint'

        Parameters
        ----------
            endpoint: Nasse.models.Endpoint
                The request's endpoint
        """
        dynamics = dynamics or {}

        if not isinstance(endpoint, models.Endpoint):
            raise exceptions.request.MissingEndpoint("The current request does not have any Nasse endpoint")

        self.nasse = app
        self.app = self.nasse
        self.nasse_endpoint = endpoint

        self.client_ip = utils.ip.get_ip()

        self.method = flask.request.method.upper()

        # sanitize
        if self.app.config.sanitize_user_input:
            self.values = werkzeug.datastructures.MultiDict((key, utils.sanitize.sanitize_text(value))
                                                            for key, value in flask.request.values.items(multi=True))
            # values.append((key, value.replace("<", "&lt").replace(">", "&gt")))
        else:
            self.values = werkzeug.datastructures.MultiDict(flask.request.values.items(multi=True))
        self.params = self.values

        if self.app.config.sanitize_user_input:
            self.args = werkzeug.datastructures.MultiDict((key, utils.sanitize.sanitize_text(value))
                                                          for key, value in flask.request.args.items(multi=True))
        else:
            self.args = werkzeug.datastructures.MultiDict(
                flask.request.args.items(multi=True))

        if self.app.config.sanitize_user_input:
            self.form = werkzeug.datastructures.MultiDict((key, utils.sanitize.sanitize_text(value))
                                                          for key, value in flask.request.form.items(multi=True))
        else:
            self.form = werkzeug.datastructures.MultiDict(
                flask.request.form.items(multi=True))

        if self.app.config.sanitize_user_input:
            self.dynamics = werkzeug.datastructures.MultiDict((key, utils.sanitize.sanitize_text(value))
                                                              for key, value in dynamics.items())
        else:
            self.dynamics = werkzeug.datastructures.MultiDict(dynamics.items())

        self.headers = werkzeug.datastructures.MultiDict(flask.request.headers)
        self.cookies = werkzeug.datastructures.MultiDict(flask.request.cookies)

        # verify if missing
        for attr, exception, current_values in [("parameters", exceptions.request.MissingParam, self.values),
                                                ("headers", exceptions.request.MissingHeader, self.headers),
                                                ("cookies", exceptions.request.MissingCookie, self.cookies),
                                                ("dynamics", exceptions.request.MissingDynamic, self.dynamics)]:
            # data: models.FinalMethodVariant[models.FinalIterable[models.UserSent]] = self.nasse_endpoint[attr]

            for value in models.get_method_variant(self.method, self.nasse_endpoint[attr]):
                if value.name not in current_values:
                    if value.required:
                        raise exception(name=value.name)
                else:
                    if value.type:
                        if callable(value.type):
                            cast = value.type
                        else:
                            cast = value.type.__class__
                        results = []
                        for key, val in current_values.items(multi=True):
                            if key == value.name:
                                try:
                                    results.append(cast(val))
                                except (ValueError, TypeError) as err:
                                    raise exceptions.request.InvalidType(name=key) from err
                        current_values.setlist(value.name, results)

    def __setattr__(self, name: str, value: typing.Any) -> None:
        if name in _overwritten:
            return super().__setattr__(name, value)
        return flask.request.__setattr__(name, value)

    def __getattribute__(self, name: str) -> typing.Any:
        if name in _overwritten:
            return super().__getattribute__(name)
        return flask.request._get_current_object().__getattribute__(name)



"""
Internal Nasse error about arguments
"""

import typing
from nasse.exceptions.base import NasseException


class MissingArgument(NasseException):
    """When an argument is missing"""
    STATUS_CODE = 500
    MESSAGE = "An argument is missing"
    EXCEPTION_NAME = "MISSING_INTERNAL_ARGUMENT"



"""Exposing the logging API"""

# pylint: disable=unused-import
from nasse.utils.logging import (CallStackRecorder, Logger, LoggingLevel,
                                 Record, StackFrame, debug, error, hidden,
                                 hide, info, log, logger, warn, warning)



"""
The english translation

Copyright
---------
Animenosekai
    Original Author, MIT License
"""

from nasse.localization.base import Localization


class EnglishLocalization(Localization):
    # it's the same one as the base
    __native__ = "English"



"""
Defines the base exception
"""

import typing
from nasse.utils.logging import logger


class NasseException(Exception):
    """
    Represents an exception, providing a bit more context for the server to actually give back the best answers possible.

    Here are the different constants which can be set

    - STATUS_CODE: The actual status code to respond with, defaults to 500
    - MESSAGE: A message accompanying the error, optional
    - EXCEPTION_NAME: Explicitely giving the exception name, defaults to "SERVER_ERROR" 
    - LOG: If the error should be logged to the console, defaults to True
    """
    STATUS_CODE = 500
    MESSAGE = "An unexpected error occured on the server"
    EXCEPTION_NAME = "SERVER_ERROR"
    LOG = True

    def __init__(self, *args: object, message: typing.Optional[str] = None) -> None:
        if message is not None:
            self.MESSAGE = str(message)

        super().__init__(self.MESSAGE, *args)

        if self.LOG:
            # should be the context's logger
            logger.error(self.MESSAGE)



"""The quitting screen"""

from textual.containers import Grid
from textual.screen import ModalScreen
from textual.widgets import Button, Label
from textual import events
import typing
from nasse.localization import Localization, EnglishLocalization


class QuitScreen(ModalScreen):
    """Screen with a dialog to quit."""

    DEFAULT_CSS = """
    QuitScreen {
        align: center middle;
    }

    #dialog {
        grid-size: 2;
        grid-gutter: 1 2;
        grid-rows: 1fr 3;
        padding: 0 1;
        width: 60;
        height: 11;
        border: thick $background 80%;
        background: $surface;
    }

    #question {
        column-span: 2;
        height: 1fr;
        width: 1fr;
        content-align: center middle;
    }

    Button {
        width: 100%;
    }
    """

    def __init__(self, localization: typing.Type[Localization] = EnglishLocalization, **kwargs) -> None:
        super().__init__(**kwargs)
        self.localization = localization

    def compose(self):
        yield Grid(
            Label(self.localization.tui_quit_confirmation, id="question"),
            Button(self.localization.tui_quit, variant="error", id="quit"),
            Button(self.localization.tui_cancel, variant="primary", id="cancel"),
            id="dialog",
        )

    def on_button_pressed(self, event: Button.Pressed) -> None:
        """When a button is pressed"""
        if event.button.id == "quit":
            self.app.exit()
        else:
            self.app.pop_screen()

    def on_key(self, event: events.Key) -> None:
        """When a key is pressed on the keyboard"""
        if event.key == "escape":
            self.app.pop_screen()



"""
Runs the Nasse instances directly from the command line

TODO
----
- Running view
- Configurations view
- Statistical Monitoring
"""

from nasse import Nasse
from nasse.tui.app import App


def headless(app: Nasse, **kwargs):
    """The headless, non interactive version of the runner"""
    app.run(**kwargs)


class Runner(App):
    """The runner app"""
    pass



"""
This folder contains the different HTTP WSGI server backends
"""
import abc
from nasse import config


class ServerBackend(abc.ABC):
    """
    A server backend
    """

    def __init__(self, app: "Nasse", config: config.NasseConfig) -> None:
        self.app = app
        self.config = config or config.NasseConfig()

    @abc.abstractmethod
    def run(self, *args, **kwargs):
        """
        Should contain the server running logic
        """
        return None

    @abc.abstractmethod
    def stop(self):
        """
        Should contain the server termination logic
        """



"""
Generates docs for Postman
"""

import typing
from copy import deepcopy
# from uuid import uuid4

from nasse import docs, models
from nasse.localization.base import Localization
from nasse.utils.sanitize import sort_http_methods


def create_postman_data(app, section: str, endpoints: typing.List[models.Endpoint], localization: Localization = Localization()) -> typing.Dict[str, typing.Any]:
    """
    Generates the data for postman

    Parameters
    ----------
    app
    section: str
        The section to document
    endpoint: list[models.Endpoint]
        A list of endpoints
    localization: Localization
        The language of the docs

    Returns
    -------
    dict[str, Any]
        The data for Postman
    """
    postman_section = {
        "info": {
            # "_postman_id": str(uuid4()),
            "name": section,
            "description": localization.postman_description.format(section=section, name=app.config.name),
            "schema": "https://schema.getpostman.com/json/collection/v2.1.0/collection.json"
        },
        "item": [],
        "auth": {
            "type": "apikey",
            "apikey": [
                {
                    "key": "value",
                    "value": "{{{{{id}_TOKEN}}}}".format(id=app.config.id.upper()),
                    "type": "string"
                },
                {
                    "key": "key",
                    "value": "Authorization",
                    "type": "string"
                }
            ]
        }
    }
    for endpoint in endpoints:
        postman_section["item"].extend(create_postman_docs(endpoint, localization=localization))
    return postman_section


def create_postman_docs(endpoint: models.Endpoint, localization: Localization = Localization()):
    """
    Generates the Postman docs for an endpoint

    Parameters
    ----------
    endpoint: models.Endpoint
        The endpoint to generate the docs for
    localization: Localization
        The language to use

    Returns
    -------
    list[dict[str, Any]]
        A list of documentation data for Postman (one item per method)
    """
    results = []
    for method in sort_http_methods(endpoint.methods):
        result = {
            "name": str(endpoint.name),
            "event": [],
            "request": {
                "method": str(method).upper(),
                "header": [
                    {
                        "key": header.name,
                        "value": header.description or header.name,
                        "type": "text"
                    }
                    for header in models.get_method_variant(method, endpoint.headers)],
                "url": {
                    "raw": "{{DOMAIN}}" + endpoint.path.replace("<", "{{").replace(">", "}}") + "?=" + '&'.join([param.name for param in models.get_method_variant(method, endpoint.parameters)]),
                    "host": [
                        "{{DOMAIN}}"
                    ],
                    "path": [elem.replace("<", "{{").replace(">", "}}") for elem in endpoint.path.split("/") if elem != ""],
                    "query": [
                        {
                            "key": param.name,
                            "value": "<{param}:{type}>".format(param=param.name, type=param.type.__name__ if hasattr(param.type, "__name__") else str(param.type) if param.type is not None else "str"),
                            "description": param.description or param.name
                        }
                        for param in models.get_method_variant(method, endpoint.parameters)]
                },
                "description": docs.markdown.make_docs_for_method(endpoint=endpoint, method=method, postman=True, localization=localization)
            },
            "response": []
        }
        result["response"].append(deepcopy(result))
        result["response"][0]["status"] = "OK"
        result["response"][0]["code"] = 200
        if endpoint.json:
            result["response"][0]["_postman_previewlanguage"] = "json"
        result["response"][0]["header"] = []
        result["response"][0]["cookie"] = []
        result["response"][0]["body"] = docs.example.generate_example(
            endpoint=endpoint, method=method)

        login_rules = endpoint.login.get(method, endpoint.login["*"])

        for rule in login_rules:
            if rule.skip:
                result["request"]["auth"] = {
                    "type": "noauth"
                }
                break
        results.append(result)

    return results
